"""Solar ordinance structured parsing class"""

import asyncio
import logging
from copy import deepcopy
from itertools import chain
from warnings import warn

import pandas as pd

from compass.llm.calling import BaseLLMCaller, ChatLLMCaller
from compass.extraction.features import SetbackFeatures
from compass.common import (
    EXTRACT_ORIGINAL_TEXT_PROMPT,
    SYSTEM_SIZE_REMINDER,
    run_async_tree,
    run_async_tree_with_bm,
    empty_output,
    setup_async_decision_tree,
    setup_base_setback_graph,
    setup_participating_owner,
    setup_graph_extra_restriction,
    setup_graph_permitted_use_districts,
)
from compass.extraction.solar.graphs import (
    setup_graph_sef_types,
    setup_multiplier,
)
from compass.utilities.enums import LLMUsageCategory
from compass.warn import COMPASSWarning
from compass.pb import COMPASS_PB


logger = logging.getLogger(__name__)
DEFAULT_SYSTEM_MESSAGE = (
    "You are a legal scholar informing a solar energy developer about local "
    "zoning ordinances."
)
SETBACKS_SYSTEM_MESSAGE = (
    f"{DEFAULT_SYSTEM_MESSAGE} "
    "For the duration of this conversation, only focus on ordinances"
    "relating to setbacks from {feature}; do not respond based on any text "
    "related to {ignore_features}. "
    f"Please only consider ordinances for {SYSTEM_SIZE_REMINDER}"
)
RESTRICTIONS_SYSTEM_MESSAGE = (
    f"{DEFAULT_SYSTEM_MESSAGE} "
    "For the duration of this conversation, only focus on "
    "ordinances relating to {restriction} for "
    f"{SYSTEM_SIZE_REMINDER}"
)
PERMITTED_USE_SYSTEM_MESSAGE = (
    f"{DEFAULT_SYSTEM_MESSAGE} "
    "For the duration of this conversation, only focus on permitted uses for "
    f"{SYSTEM_SIZE_REMINDER}"
)
EXTRA_NUMERICAL_RESTRICTIONS = {
    "other secs": (
        "**minimum** required separation distance with other existing or "
        "planned solar energy conversion systems"
    ),
    "noise": "maximum noise level allowed",
    "maximum height": "maximum structure height allowed",
    "maximum project size": (
        "maximum project size or total installation allowed"
    ),
    "minimum lot size": "**minimum** lot, parcel, or tract size allowed",
    "maximum lot size": "**maximum** lot, parcel, or tract size allowed",
    "panel spacing": (
        "**minimum** allowed spacing between individual solar panels"
    ),
    "land density": "**maximum** allowed system size **per unit area**",
    "coverage": "**maximum** land coverage allowed",
}
EXTRA_QUALITATIVE_RESTRICTIONS = {
    "prohibitions": "prohibitions, moratoria, or bans",
    "decommissioning": "decommissioning requirements",
    "glare": "glare restrictions",
    "visual impact": "visual impact **assessment** requirements",
    "repowering": (
        "requirements or regulations specific to **repowering** of the system"
    ),
    "fencing": "fencing requirements",
    "signage": "signage requirements",
    "screening": "vegetative buffer and/or other screening requirements",
    "soil": "soil, erosion, and/or sediment control requirements",
}
UNIT_CLARIFICATIONS = {
    "noise": (
        "For the purposes of this extraction, assume the standard units "
        "for noise are 'dBA'."
    ),
    "panel spacing": (
        "For the purposes of this extraction, assume the standard units for"
        "spacing between individual solar panels are one of the following: "
        "'height-multiplier', 'feet', or 'meters'."
    ),
    "land density": (
        "This value should be specified as a maximum system "
        "size value (in MW) per area (square meters, acres, etc.). "
        "Do not confuse this with lot coverage or min/max lot size "
        "requirements."
    ),
    "coverage": (
        "Maximum land coverage should be specified as a fraction or "
        "percentage of the total land."
    ),
    "minimum lot size": (
        "Minimum lot size should **always** be specified as an area value."
    ),
    "maximum lot size": (
        "Maximum lot size should **always** be specified as an area value."
    ),
}
ER_CLARIFICATIONS = {
    "maximum project size": (
        "Maximum project size should be specified as a maximum system "
        "size value (in MW) or as a limit on installation size (e.g. maximum "
        "number of systems or maximum number of solar panels). "
        "Do not confuse this with lot coverage or min/max lot size "
        "requirements."
    ),
    "land density": (
        "Do **not** try to infer the density requirement based on other "
        "restrictions such as setbacks from facility perimeters, property "
        "lines, etc."
    ),
    "visual impact": (
        "Do **not** consider glare restrictions as part of visual impact - "
        "these are collected separately."
    ),
    "panel spacing": (
        "Do **not** try to infer the spacing requirement based on other "
        "restrictions such as setbacks from facility perimeters, property "
        "lines, etc."
    ),
    "coverage": (
        "Do not confuse this with density or min/max lot size requirements."
    ),
}
_FEATURE_TO_OWNED_TYPE = {
    "structures": "structure",
    "property line": "property",
}


class StructuredSolarParser(BaseLLMCaller):
    """Base class for parsing structured data"""

    def _init_chat_llm_caller(self, system_message):
        """Initialize a ChatLLMCaller instance for the DecisionTree"""
        return ChatLLMCaller(
            self.llm_service,
            system_message=system_message,
            usage_tracker=self.usage_tracker,
            **self.kwargs,
        )

    async def _check_solar_farm_type(self, text):
        """Get the largest solar farm size mentioned in the text"""
        logger.info("Checking solar farm types")
        tree = setup_async_decision_tree(
            setup_graph_sef_types,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(DEFAULT_SYSTEM_MESSAGE),
        )
        decision_tree_sef_types_out = await run_async_tree(tree)
        largest_system = (
            decision_tree_sef_types_out.get("largest_sef_type")
            or "utility-scale solar energy systems"
        )

        if not decision_tree_sef_types_out.get("is_large", True):
            logger.info(
                "Did not find utility-scale systems in text. Largest "
                "system found: %r",
                largest_system,
            )
            return None

        logger.info("Largest SEF type found in text: %r", largest_system)
        return largest_system


class StructuredSolarOrdinanceParser(StructuredSolarParser):
    """LLM ordinance document structured data scraping utility

    Purpose:
        Extract structured ordinance data from text.
    Responsibilities:
        1. Extract ordinance values into structured format by executing
           a decision-tree-based chain-of-thought prompt on the text for
           each value to be extracted.
    Key Relationships:
        Uses a StructuredLLMCaller for LLM queries and multiple
        AsyncDecisionTree instances to guide the extraction of
        individual values.
    """

    async def parse(self, text):
        """Parse text and extract structure ordinance data

        Parameters
        ----------
        text : str
            Ordinance text which may or may not contain setbacks for one
            or more features (property lines, structure, roads, etc.).
            Text can also contain other supported regulations (noise,
            shadow-flicker, etc,) which will be extracted as well.

        Returns
        -------
        pandas.DataFrame or None
            DataFrame containing parsed-out ordinance values. Can also
            be ``None`` if a large solar energy system is not found in
            the text.
        """
        largest_sef_type = await self._check_solar_farm_type(text)
        if not largest_sef_type:
            return None

        outer_task_name = asyncio.current_task().get_name()
        num_to_process = (
            len(SetbackFeatures.DEFAULT_FEATURE_DESCRIPTIONS)
            + len(EXTRA_NUMERICAL_RESTRICTIONS)
            + len(EXTRA_QUALITATIVE_RESTRICTIONS)
        )
        with COMPASS_PB.jurisdiction_sub_prog_bar(outer_task_name) as sub_pb:
            task_id = sub_pb.add_task(
                "Extracting ordinance values...",
                total=num_to_process,
                just_parsed="",
            )
            outputs = await self._parse_all_restrictions_with_pb(
                sub_pb, task_id, text, largest_sef_type, outer_task_name
            )
            sub_pb.update(task_id, completed=num_to_process)
            await asyncio.sleep(1)
            sub_pb.remove_task(task_id)

        return pd.DataFrame(chain.from_iterable(outputs))

    async def _parse_all_restrictions_with_pb(
        self, sub_pb, task_id, text, largest_sef_type, outer_task_name
    ):
        """Parse all ordinance values"""
        feature_parsers = [
            asyncio.create_task(
                self._parse_setback_feature(
                    sub_pb, task_id, text, feature_kwargs, largest_sef_type
                ),
                name=outer_task_name,
            )
            for feature_kwargs in SetbackFeatures()
        ]
        extras_parsers = [
            asyncio.create_task(
                self._parse_extra_restriction(
                    sub_pb,
                    task_id,
                    text,
                    feature_id,
                    r_text,
                    largest_sef_type,
                    is_numerical=True,
                    unit_clarification=UNIT_CLARIFICATIONS.get(feature_id, ""),
                    feature_clarifications=ER_CLARIFICATIONS.get(
                        feature_id, ""
                    ),
                ),
                name=outer_task_name,
            )
            for feature_id, r_text in EXTRA_NUMERICAL_RESTRICTIONS.items()
        ]
        extras_parsers += [
            asyncio.create_task(
                self._parse_extra_restriction(
                    sub_pb,
                    task_id,
                    text,
                    feature_id,
                    r_text,
                    largest_sef_type,
                    is_numerical=False,
                    feature_clarifications=ER_CLARIFICATIONS.get(
                        feature_id, ""
                    ),
                ),
                name=outer_task_name,
            )
            for feature_id, r_text in EXTRA_QUALITATIVE_RESTRICTIONS.items()
        ]
        return await asyncio.gather(*(feature_parsers + extras_parsers))

    async def _parse_extra_restriction(
        self,
        sub_pb,
        task_id,
        text,
        feature_id,
        restriction_text,
        largest_sef_type,
        is_numerical,
        unit_clarification="",
        feature_clarifications="",
    ):
        """Parse a non-setback restriction from the text"""
        logger.debug("Parsing extra feature %r", feature_id)
        system_message = RESTRICTIONS_SYSTEM_MESSAGE.format(
            restriction=restriction_text, tech=largest_sef_type
        )
        tree = setup_async_decision_tree(
            setup_graph_extra_restriction,
            usage_sub_label=LLMUsageCategory.ORDINANCE_VALUE_EXTRACTION,
            is_numerical=is_numerical,
            tech=largest_sef_type,
            feature_id=feature_id,
            restriction=restriction_text,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
            unit_clarification=unit_clarification,
            feature_clarifications=feature_clarifications,
        )
        info = await run_async_tree(tree)
        info.update({"feature": feature_id, "quantitative": is_numerical})
        if is_numerical:
            info = _sanitize_output(info)
        sub_pb.update(task_id, advance=1, just_parsed=feature_id)
        return [info]

    async def _parse_setback_feature(
        self, sub_pb, task_id, text, feature_kwargs, largest_sef_type
    ):
        """Parse values for a setback feature"""
        feature_id = feature_kwargs["feature_id"]
        feature_kwargs["tech"] = largest_sef_type
        logger.debug("Parsing feature %r", feature_id)

        out, base_messages = await self._base_messages(text, **feature_kwargs)
        if not out:
            logger.debug("Did not find ordinance for feature %r", feature_id)
            sub_pb.update(task_id, advance=1, just_parsed=feature_id)
            return empty_output(feature_id)

        if feature_id not in {"structures", "property line"}:
            output = {"feature": feature_id}
            output.update(
                await self._extract_setback_values(
                    text=text,
                    base_messages=base_messages,
                    **feature_kwargs,
                )
            )
            sub_pb.update(task_id, advance=1, just_parsed=feature_id)
            return [output]

        output = await self._extract_setback_values_for_p_or_np(
            text, base_messages, **feature_kwargs
        )
        sub_pb.update(task_id, advance=1, just_parsed=feature_id)
        return output

    async def _base_messages(self, text, **feature_kwargs):
        """Get base messages for setback feature parsing"""
        system_message = SETBACKS_SYSTEM_MESSAGE.format(**feature_kwargs)
        tree = setup_async_decision_tree(
            setup_base_setback_graph,
            usage_sub_label=LLMUsageCategory.ORDINANCE_VALUE_EXTRACTION,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
            **feature_kwargs,
        )
        out = await run_async_tree(tree, response_as_json=False)
        return out, deepcopy(tree.chat_llm_caller.messages)

    async def _extract_setback_values_for_p_or_np(
        self, text, base_messages, feature_id, **feature_kwargs
    ):
        """Extract setback values for participating ordinances"""
        logger.debug("Checking participating vs non-participating")
        p_np_text = {"participating": "", "non-participating": text}
        decision_tree_participating_out = await self._run_setback_graph(
            setup_participating_owner,
            text,
            base_messages=deepcopy(base_messages),
            owned_type=_FEATURE_TO_OWNED_TYPE[feature_id],
            **feature_kwargs,
        )
        p_np_text.update(decision_tree_participating_out)
        outer_task_name = asyncio.current_task().get_name()
        p_or_np_parsers = [
            asyncio.create_task(
                self._parse_p_or_np_text(
                    key, sub_text, base_messages, feature_id, **feature_kwargs
                ),
                name=outer_task_name,
            )
            for key, sub_text in p_np_text.items()
        ]
        return await asyncio.gather(*p_or_np_parsers)

    async def _parse_p_or_np_text(
        self, p_or_np, sub_text, base_messages, feature_id, **feature_kwargs
    ):
        """Parse participating sub-text for ord values"""
        out_feat_name = f"{feature_id} ({p_or_np})"
        output = {"feature": out_feat_name}
        if not sub_text:
            return output

        feature = feature_kwargs["feature"]
        if (
            p_or_np == "participating"
            or "non-participating"
            in sub_text.casefold().replace("\n", "").replace(" ", "-")
        ):
            feature = f"**{p_or_np}** {feature}"
            feature_kwargs["feature"] = feature

        base_messages = deepcopy(base_messages)
        base_messages[-2]["content"] = EXTRACT_ORIGINAL_TEXT_PROMPT.format(
            feature=feature,
            tech=feature_kwargs["tech"],
            ignore_features=feature_kwargs["ignore_features"],
            feature_clarifications=feature_kwargs.get(
                "feature_clarifications", ""
            ),
        )
        base_messages[-1]["content"] = sub_text

        values = await self._extract_setback_values(
            text=sub_text,
            base_messages=deepcopy(base_messages),
            **feature_kwargs,
        )
        output.update(values)
        return output

    async def _extract_setback_values(self, **kwargs):
        """Extract setback values for a given feature from input text"""
        decision_tree_out = await self._run_setback_graph(
            setup_multiplier, **kwargs
        )
        decision_tree_out = _update_output_keys(decision_tree_out)
        return _sanitize_output(decision_tree_out)

    async def _run_setback_graph(
        self, graphs_setup_func, text, base_messages=None, **kwargs
    ):
        """Generic function to run async tree"""
        system_message = SETBACKS_SYSTEM_MESSAGE.format(**kwargs)
        tree = setup_async_decision_tree(
            graphs_setup_func,
            usage_sub_label=LLMUsageCategory.ORDINANCE_VALUE_EXTRACTION,
            text=text,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
            **kwargs,
        )
        if base_messages:
            return await run_async_tree_with_bm(tree, base_messages)
        return await run_async_tree(tree)


class StructuredSolarPermittedUseDistrictsParser(StructuredSolarParser):
    """LLM permitted use districts scraping utility

    Purpose:
        Extract structured ordinance data from text.
    Responsibilities:
        1. Extract ordinance values into structured format by executing
           a decision-tree-based chain-of-thought prompt on the text for
           each value to be extracted.
    Key Relationships:
        Uses a StructuredLLMCaller for LLM queries and multiple
        AsyncDecisionTree instances to guide the extraction of
        individual values.
    """

    _LARGE_SEF_CLARIFICATION = (
        "Large solar energy systems (SES) may also be referred to as solar "
        "panels, solar energy conversion systems (SECS), solar energy "
        "facilities (SEF), solar energy farms (SEF), solar farms (SF), "
        "utility-scale solar energy systems (USES), commercial solar energy "
        "systems (CSES), alternate energy systems (AES), commercial energy "
        "production systems (CEPCS), or similar"
    )
    _USE_TYPES = [
        {
            "feature_id": "primary use districts",
            "use_type": (
                "permitted as primary use or similar (e.g., without special "
                "conditions or approval)"
            ),
            "clarifications": (
                "Consider any solar overlay districts as "
                "primary use districts. {sef_clarification}"
            ),
        },
        {
            "feature_id": "special use districts",
            "use_type": (
                "permitted as special use or similar (e.g., requires approval "
                "by the zoning appeals board or meeting certain conditions "
                "like completing a permitting process)"
            ),
            "clarifications": (
                "Consider any solar overlay districts as "
                "primary use and **do not include** them in the output. "
                "{sef_clarification}"
            ),
        },
        {
            "feature_id": "accessory use districts",
            "use_type": (
                "permitted as accessory use or similar (e.g., when integrated "
                "with an existing structure or secondary to another use)"
            ),
            "clarifications": (
                "Consider any solar overlay districts as "
                "primary use and **do not include** them in the output. "
                "{sef_clarification}"
            ),
        },
        {
            "feature_id": "prohibited use districts",
            "use_type": (
                "prohibited or similar (e.g., where solar energy "
                "systems are not allowed or banned)"
            ),
            "clarifications": (
                "Only output specific districts where solar energy systems "
                "are prohibited **unconditionally**. "
                "{sef_clarification}"
            ),
        },
    ]

    async def parse(self, text):
        """Parse text and extract permitted use districts data

        Parameters
        ----------
        text : str
            Permitted use districts text which may or may not contain
            information about allowed uses in one or more districts.

        Returns
        -------
        pandas.DataFrame or None
            DataFrame containing parsed-out allowed-use district names.
            Can also be ``None`` if a large solar energy system is not
            found in the text.
        """
        largest_sef_type = await self._check_solar_farm_type(text)
        if not largest_sef_type:
            return None

        loc = asyncio.current_task().get_name()
        with COMPASS_PB.jurisdiction_sub_prog_bar(loc) as sub_pb:
            task_id = sub_pb.add_task(
                "Extracting permitted uses...",
                total=len(self._USE_TYPES),
                just_parsed="",
            )
            feature_parsers = [
                asyncio.create_task(
                    self._parse_permitted_use_districts(
                        sub_pb,
                        task_id,
                        text,
                        largest_sef_type,
                        **use_type_kwargs,
                    ),
                    name=loc,
                )
                for use_type_kwargs in self._USE_TYPES
            ]
            outputs = await asyncio.gather(*(feature_parsers))
            sub_pb.update(task_id, completed=len(self._USE_TYPES))
            await asyncio.sleep(1)
            sub_pb.remove_task(task_id)

        return pd.DataFrame(chain.from_iterable(outputs))

    async def _parse_permitted_use_districts(
        self,
        sub_pb,
        task_id,
        text,
        largest_sef_type,
        feature_id,
        use_type,
        clarifications,
    ):
        """Parse a non-setback restriction from the text"""
        logger.debug("Parsing use type: %r", feature_id)
        system_message = PERMITTED_USE_SYSTEM_MESSAGE.format(
            tech=largest_sef_type
        )
        tree = setup_async_decision_tree(
            setup_graph_permitted_use_districts,
            usage_sub_label=LLMUsageCategory.PERMITTED_USE_VALUE_EXTRACTION,
            feature_id=feature_id,
            tech=largest_sef_type,
            clarifications=clarifications.format(
                sef_clarification=self._LARGE_SEF_CLARIFICATION
            ),
            text=text,
            use_type=use_type,
            chat_llm_caller=self._init_chat_llm_caller(system_message),
        )
        info = await run_async_tree(tree)
        sub_pb.update(task_id, advance=1, just_parsed=feature_id)
        info.update({"feature": feature_id, "quantitative": True})
        return [info]


def _update_output_keys(output):
    """Standardize output keys

    We could standardize output keys by modifying the LLM prompts, but
    have found that it's more accurate to instruct the LLM to use
    descriptive keys (e.g. "mult_value" instead of "value" or
    "mult_type" instead of "units")
    """

    if "mult_value" not in output:
        return output

    output["value"] = output.pop("mult_value")

    if units := output.get("units"):
        msg = f"Found non-null units value for multiplier: {units}"
        warn(msg, COMPASSWarning)
    output["units"] = "structure-height-multiplier"

    return output


def _sanitize_output(output):
    """Perform some sanitization on outputs"""
    output = _remove_key_for_empty_value(output, key="units")
    return _remove_key_for_empty_value(output, key="summary")


def _remove_key_for_empty_value(output, key):
    """Remove any output in "key" if no ordinance value found"""
    if output.get("value") is not None or not output.get(key):
        return output

    # at this point, we have some value in "key" but no actual ordinance
    # value, so remove the "key" entry
    output[key] = None
    return output



"""Solar ordinance decision tree graph setup functions"""

from compass.common import (
    setup_graph_no_nodes,
    llm_response_starts_with_yes,
    llm_response_starts_with_no,
    SYSTEM_SIZE_REMINDER,
)


def setup_graph_sef_types(**kwargs):
    """Setup graph to get the largest solar farm size in the text

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    networkx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(  # noqa: N806
        d_tree_name="Solar Energy Farm types", **kwargs
    )

    G.add_node(
        "init",
        prompt=(
            "Does the following text distinguish between multiple solar "
            "energy farm sizes? Distinctions are often made as 'small', "
            "'personal', or 'private' vs 'large', 'commercial', or 'utility'. "
            "Sometimes the distinction uses actual MW values. "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain your answer."
            '\n\n"""\n{text}\n"""'
        ),
    )

    G.add_edge("init", "get_text", condition=llm_response_starts_with_yes)
    G.add_node(
        "get_text",
        prompt=(
            "What are the different solar energy farm sizes **regulated by "
            "this ordinance**? List them in order of increasing size. "
            "Include any relevant numerical qualifiers in the name, if "
            "appropriate. Only include systems that resemble ground-mounted "
            "solar energy farms; do not include other solar energy system "
            "types lice CSP or roof-mounted systems or other technologies "
            "like wind energy systems, geothermal energy systems, etc."
        ),
    )
    G.add_edge("get_text", "get_regulated")
    G.add_node(
        "get_regulated",
        prompt=(
            "Are any of these systems **not** regulated by this ordinance?"
        ),
    )

    G.add_edge("get_regulated", "get_largest")
    G.add_node(
        "get_largest",
        prompt=(
            "What is the **largest** solar energy farm size that **is "
            "regulated by this ordinance**?"
        ),
    )

    G.add_edge("get_largest", "check_matches_definition")
    G.add_node(
        "check_matches_definition",
        prompt=(
            "Does the ordinance explicitly define this system as large, "
            "commercial, utility-scale, or something akin to that? "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain your answer."
        ),
    )

    G.add_edge(
        "check_matches_definition",
        "final_large",
        condition=llm_response_starts_with_yes,
    )
    G.add_edge(
        "check_matches_definition",
        "check_scale_reason",
        condition=llm_response_starts_with_no,
    )
    G.add_node(
        "check_scale_reason",
        prompt=(
            "Would a reasonable person classify this kind of system as a "
            "**large**, commercial, or even utility-scale** solar energy farm "
            "(e.g. with the primary purpose of generating electricity for "
            "sale, as opposed to small, residential, roof-mounted, private, "
            "or other kinds of 'small' systems)? "
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain your answer."
        ),
    )

    G.add_edge(
        "check_scale_reason",
        "final_large",
        condition=llm_response_starts_with_yes,
    )
    G.add_edge(
        "check_scale_reason",
        "final_small",
        condition=llm_response_starts_with_no,
    )
    G.add_node(
        "final_large",
        prompt=(
            "Respond based on our entire conversation so far. Return your "
            "answer as a dictionary in JSON format (not markdown). Your "
            "JSON file must include exactly two keys. The keys are "
            "'largest_sef_type' and 'explanation', and 'is_large'. The value "
            "of the 'largest_sef_type' key should be a string that labels the "
            "largest solar energy system size **regulated by this "
            "ordinance**. The value of the 'explanation' key should be a "
            "string containing a short explanation for your choice. The value "
            "of the 'is_large' key should be the boolean value `true`, since "
            "we determined this is a large-scale system."
        ),
    )
    G.add_node(
        "final_small",
        prompt=(
            "Respond based on our entire conversation so far. Return your "
            "answer as a dictionary in JSON format (not markdown). Your "
            "JSON file must include exactly two keys. The keys are "
            "'largest_sef_type' and 'explanation', and 'is_large'. The value "
            "of the 'largest_sef_type' key should be a string that labels the "
            "largest solar energy system size **regulated by this "
            "ordinance**. The value of the 'explanation' key should be a "
            "string containing a short explanation for your choice. The value "
            "of the 'is_large' key should be the boolean value `false`, since "
            "we determined this is not a large-scale system."
        ),
    )
    return G


def setup_multiplier(**kwargs):
    """Setup graph to extract a setbacks multiplier values for a feature

    Parameters
    ----------
    **kwargs
        Keyword-value pairs to add to graph.

    Returns
    -------
    networkx.DiGraph
        Graph instance that can be used to initialize an
        `elm.tree.DecisionTree`.
    """
    G = setup_graph_no_nodes(  # noqa: N806
        d_tree_name="Setback distance", **kwargs
    )

    G.add_node(
        "init",
        prompt=(
            "Does the text mention a multiplier that should be applied to the "
            "structure height to compute the setback distance from {feature} "
            "for {tech}? "
            "Please consider only {feature}; do not respond based on any text "
            "related to {ignore_features}. "
            "Please also only consider setbacks specifically for "
            f"{SYSTEM_SIZE_REMINDER}"
            "Please start your response with either 'Yes' or 'No' and briefly "
            "explain your answer."
        ),
    )
    G.add_edge("init", "no_multiplier", condition=llm_response_starts_with_no)
    G.add_node(
        "no_multiplier",
        prompt=(
            "Does the ordinance give the setback from {feature} as a fixed "
            "distance value? "
            "Please consider only on {feature}; do not respond based on any "
            "text related to {ignore_features}. "
            "Please also only consider setbacks specifically for "
            f"{SYSTEM_SIZE_REMINDER}"
            "Please start your response with either 'Yes' or "
            "'No' and briefly explain your answer."
        ),
    )
    G.add_edge(
        "no_multiplier", "units", condition=llm_response_starts_with_yes
    )
    G.add_edge(
        "no_multiplier", "out_static", condition=llm_response_starts_with_no
    )
    G.add_node(
        "units",
        prompt=(
            "What are the units for the setback from {feature}? "
            "Ensure that:\n1) You accurately identify the unit value "
            "associated with the setback.\n2) The unit is "
            "expressed using standard, conventional unit names (e.g., "
            "'feet', 'meters', 'miles' etc.)\n3) If multiple "
            "values are mentioned, return only the units for the most "
            "restrictive value that directly pertains to the setback.\n\n"
            "Example Inputs and Outputs:\n"
            "Text: 'All WECS Towers shall be set back a distance of at least "
            "one thousand (1000) feet, from any primary structure'\n"
            "Output: 'feet'\n"
        ),
    )
    G.add_edge("units", "out_static")
    G.add_node(
        "out_static",
        prompt=(
            "Please respond based on our entire conversation so far. "
            "Return your answer in JSON "
            "format (not markdown). Your JSON file must include exactly "
            "four keys. The keys are 'value', 'units', 'summary', and "
            "'section'. The value of the 'value' key should be a "
            "**numerical** value corresponding to the setback distance value "
            "from {feature} or `null` if there was no such value. The value "
            "of the 'units' key should be a string corresponding to the "
            "(standard) units of the setback distance value from {feature} "
            "or `null` if there was no such value. "
            "As before, focus only on setbacks that would apply for"
            f"{SYSTEM_SIZE_REMINDER}"
            "{SUMMARY_PROMPT} {SECTION_PROMPT}"
        ),
    )
    G.add_edge("init", "m_single", condition=llm_response_starts_with_yes)

    G.add_node(
        "m_single",
        prompt=(
            "Are multiple values given for the multiplier used to "
            "compute the setback distance value from {feature} for {tech}? "
            "Remember to ignore any text related to {ignore_features}. "
            "Please only consider setbacks specifically for "
            f"{SYSTEM_SIZE_REMINDER}"
            "If so, select and state the largest one. Otherwise, repeat the "
            "single multiplier value that was given in the text. "
        ),
    )
    G.add_edge("m_single", "adder")
    G.add_node(
        "adder",
        prompt=(
            "Does the ordinance for the setback from {feature} include a "
            "static distance value that should be added to the result of "
            "the multiplication? "
            "Remember to ignore any text related to {ignore_features}. "
            "Please only consider setbacks specifically for "
            f"{SYSTEM_SIZE_REMINDER}"
            "Do not confuse this value with static setback requirements. "
            "Ignore text with clauses such as 'no lesser than', 'no greater "
            "than', 'the lesser of', or 'the greater of'. Please start your "
            "response with either 'Yes' or 'No' and briefly explain your "
            "answer, stating the adder value if it exists."
        ),
    )
    G.add_edge("adder", "out_no_adder", condition=llm_response_starts_with_no)
    G.add_edge("adder", "adder_eq", condition=llm_response_starts_with_yes)

    G.add_node(
        "adder_eq",
        prompt=(
            "Does the adder value you identified satisfy the following "
            "equation: `multiplier * height + <adder>`? Begin your "
            "response with either 'Yes' or 'No' and briefly explain your "
            "answer."
        ),
    )
    G.add_edge(
        "adder_eq", "out_no_adder", condition=llm_response_starts_with_no
    )
    G.add_edge("adder_eq", "out_m", condition=llm_response_starts_with_no)
    G.add_edge(
        "adder_eq",
        "conversion",
        condition=llm_response_starts_with_yes,
    )
    G.add_node(
        "conversion",
        prompt=(
            "If the adder value is not given in feet, convert "
            "it to feet (remember that there are 3.28084 feet in one meter "
            "and 5280 feet in one mile). Show your work step-by-step "
            "if you had to perform a conversion."
        ),
    )
    G.add_edge("conversion", "out_m")
    G.add_node(
        "out_m",
        prompt=(
            "Please respond based on our entire conversation so far. "
            "Return your answer as a single dictionary in JSON "
            "format (not markdown). Your JSON file must include exactly four "
            "keys. The keys are 'mult_value', 'adder', 'summary', and "
            "'section'. The value of the 'mult_value' key should be a "
            "**numerical** value corresponding to the multiplier value we "
            "determined earlier. The value of the 'adder' key should be a "
            "**numerical** value corresponding to the static value to be "
            "added to the total setback distance after multiplication, as we "
            "determined earlier, or `null` if there is no such value. "
            "{SUMMARY_PROMPT} {SECTION_PROMPT}"
        ),
    )
    G.add_node(
        "out_no_adder",
        prompt=(
            "Please respond based on our entire conversation so far. "
            "Return your answer as a single dictionary in JSON "
            "format (not markdown). Your JSON file must include exactly three "
            "keys. The keys are 'mult_value', 'summary', and 'section'. The "
            "value of the 'mult_value' key should be a **numerical** value "
            "corresponding to the multiplier value we determined earlier. "
            "{SUMMARY_PROMPT} {SECTION_PROMPT}"
        ),
    )

    return G



"""COMPASS namedtuple data classes"""

from collections import namedtuple

ProcessKwargs = namedtuple(
    "ProcessKwargs",
    [
        "known_doc_urls",
        "file_loader_kwargs",
        "td_kwargs",
        "tpe_kwargs",
        "ppe_kwargs",
        "max_num_concurrent_jurisdictions",
    ],
    defaults=[None, None, None, None, 25],
)

TechSpec = namedtuple(
    "TechSpec",
    [
        "name",
        "questions",
        "heuristic",
        "ordinance_text_collector",
        "ordinance_text_extractor",
        "permitted_use_text_collector",
        "permitted_use_text_extractor",
        "structured_ordinance_parser",
        "structured_permitted_use_parser",
        "website_url_keyword_points",
    ],
)



"""COMPASS Ordinance jurisdiction specification utilities"""

from functools import cached_property


JURISDICTION_TYPES_AS_PREFIXES = {
    "town",
    "township",
    "city",
    "borough",
    "village",
    "unorganized territory",
}


class Jurisdiction:
    """Class representing a jurisdiction"""

    def __init__(
        self,
        subdivision_type,
        state,
        county=None,
        subdivision_name=None,
        code=None,
    ):
        """

        Parameters
        ----------
        subdivision_type : str
            Type of subdivision that this jurisdiction represents.
            Typical values are "state", "county", "town", "city",
            "borough", "parish", "township", etc.
        state : str
            Name of the state containing the jurisdiction.
        county : str, optional
            Name of the county containing the jurisdiction, if
            applicable. If the jurisdiction represents a state, leave
            this input unspecified. If the jurisdiction represents a
            county or a subdivision within a county, provide the county
            name here.

            .. IMPORTANT:: Make sure this input is capitalized properly!

            By default, ``None``.
        subdivision_name : str, optional
            Name of the subdivision that the jurisdiction represents, if
            applicable. If the jurisdiction represents a state or
            county, leave this input unspecified. Otherwise, provide the
            jurisdiction name here.

            .. IMPORTANT:: Make sure this input is capitalized properly!

            By default, ``None``.
        code : int or str, optional
            Optional jurisdiction code (typically FIPS or similar).
            By default, ``None``.
        """
        self.type = subdivision_type.title()
        self.state = state.title()
        self.county = county
        self.subdivision_name = subdivision_name
        self.code = code

    @cached_property
    def full_name(self):
        """str: Full jurisdiction name"""
        name_parts = [
            self.full_subdivision_phrase,
            self.full_county_phrase,
            self.state,
        ]

        return ", ".join(filter(None, name_parts))

    @cached_property
    def full_name_the_prefixed(self):
        """str: Full jurisdiction name with `the` prefix if needed"""
        if self.type.casefold() == "state":
            return f"the state of {self.state}"

        if self.type.casefold() in JURISDICTION_TYPES_AS_PREFIXES:
            return f"the {self.full_name}"

        return self.full_name

    @cached_property
    def full_subdivision_phrase(self):
        """str: Full jurisdiction subdivision phrase, or empty str"""
        if not self.subdivision_name:
            return ""

        if self.type.casefold() in JURISDICTION_TYPES_AS_PREFIXES:
            return f"{self.type} of {self.subdivision_name}"

        return f"{self.subdivision_name} {self.type}"

    @cached_property
    def full_subdivision_phrase_the_prefixed(self):
        """str: Full jurisdiction subdivision phrase, or empty str"""
        if self.type.casefold() in JURISDICTION_TYPES_AS_PREFIXES:
            return f"the {self.full_subdivision_phrase}"

        return self.full_subdivision_phrase

    @cached_property
    def full_county_phrase(self):
        """str: Full jurisdiction county phrase, or empty str"""
        if not self.county:
            return ""

        if not self.subdivision_name:
            return f"{self.county} {self.type}"

        return f"{self.county} County"

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.full_name

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (
                self.type.casefold() == other.type.casefold()
                and self.state.casefold() == other.state.casefold()
                and self.county == other.county
                and self.subdivision_name == other.subdivision_name
            )
        if isinstance(other, str):
            return self.full_name.casefold() == other.casefold()
        return False

    def __hash__(self):
        return hash(self.full_name.casefold())



"""Ordinance full processing logic"""

import time
import asyncio
import logging
from copy import deepcopy
from functools import cached_property
from contextlib import AsyncExitStack, contextmanager
from datetime import datetime, UTC

import pandas as pd
from elm.web.utilities import get_redirected_url

from compass import __version__
from compass.scripts.download import (
    find_jurisdiction_website,
    download_known_urls,
    download_jurisdiction_ordinance_using_search_engine,
    download_jurisdiction_ordinances_from_website,
    download_jurisdiction_ordinances_from_website_compass_crawl,
    filter_ordinance_docs,
)
from compass.exceptions import COMPASSValueError
from compass.extraction import (
    extract_ordinance_values,
    extract_ordinance_text_with_ngram_validation,
)
from compass.extraction.solar import (
    SolarHeuristic,
    SolarOrdinanceTextCollector,
    SolarOrdinanceTextExtractor,
    SolarPermittedUseDistrictsTextCollector,
    SolarPermittedUseDistrictsTextExtractor,
    StructuredSolarOrdinanceParser,
    StructuredSolarPermittedUseDistrictsParser,
    SOLAR_QUESTION_TEMPLATES,
    BEST_SOLAR_ORDINANCE_WEBSITE_URL_KEYWORDS,
)
from compass.extraction.wind import (
    WindHeuristic,
    WindOrdinanceTextCollector,
    WindOrdinanceTextExtractor,
    WindPermittedUseDistrictsTextCollector,
    WindPermittedUseDistrictsTextExtractor,
    StructuredWindOrdinanceParser,
    StructuredWindPermittedUseDistrictsParser,
    WIND_QUESTION_TEMPLATES,
    BEST_WIND_ORDINANCE_WEBSITE_URL_KEYWORDS,
)
from compass.validation.location import JurisdictionWebsiteValidator
from compass.llm import LLMCaller, OpenAIConfig
from compass.services.cpu import (
    PDFLoader,
    OCRPDFLoader,
    read_pdf_doc,
    read_pdf_doc_ocr,
)
from compass.services.usage import UsageTracker
from compass.services.openai import usage_from_response
from compass.services.provider import RunningAsyncServices
from compass.services.threaded import (
    TempFileCachePB,
    TempFileCache,
    FileMover,
    CleanedFileWriter,
    OrdDBFileWriter,
    UsageUpdater,
    JurisdictionUpdater,
)
from compass.utilities import (
    LLM_COST_REGISTRY,
    compile_run_summary_message,
    doc_infos_to_db,
    load_all_jurisdiction_info,
    load_jurisdictions_from_fp,
    num_ordinances_in_doc,
    save_db,
    save_run_meta,
    Directories,
    ProcessKwargs,
    TechSpec,
)
from compass.utilities.enums import LLMTasks
from compass.utilities.location import Jurisdiction
from compass.utilities.logs import (
    LocationFileLog,
    LogListener,
    NoLocationFilter,
)
from compass.utilities.base import WebSearchParams
from compass.utilities.parsing import load_config
from compass.pb import COMPASS_PB


logger = logging.getLogger(__name__)
EXCLUDE_FROM_ORD_DOC_CHECK = {
    # if doc only contains these, it's not good enough to count as an
    # ordinance. Note that prohibitions are explicitly not on this list
    "color",
    "decommissioning",
    "lighting",
    "visual impact",
    "glare",
    "repowering",
    "fencing",
    "climbing prevention",
    "signage",
    "soil",
    "primary use districts",
    "special use districts",
    "accessory use districts",
}
_TEXT_EXTRACTION_TASKS = {
    WindOrdinanceTextExtractor: "Extracting wind ordinance text",
    WindPermittedUseDistrictsTextExtractor: (
        "Extracting wind permitted use text"
    ),
    SolarOrdinanceTextExtractor: "Extracting solar ordinance text",
    SolarPermittedUseDistrictsTextExtractor: (
        "Extracting solar permitted use text"
    ),
}
_JUR_COLS = [
    "Jurisdiction Type",
    "State",
    "County",
    "Subdivision",
    "FIPS",
    "Website",
]
MAX_CONCURRENT_SEARCH_ENGINE_QUERIES = 10


async def process_jurisdictions_with_openai(  # noqa: PLR0917, PLR0913
    out_dir,
    tech,
    jurisdiction_fp,
    model="gpt-4o-mini",
    num_urls_to_check_per_jurisdiction=5,
    max_num_concurrent_browsers=10,
    max_num_concurrent_website_searches=10,
    max_num_concurrent_jurisdictions=25,
    url_ignore_substrings=None,
    known_doc_urls=None,
    file_loader_kwargs=None,
    search_engines=None,
    pytesseract_exe_fp=None,
    td_kwargs=None,
    tpe_kwargs=None,
    ppe_kwargs=None,
    log_dir=None,
    clean_dir=None,
    ordinance_file_dir=None,
    jurisdiction_dbs_dir=None,
    perform_website_search=True,
    llm_costs=None,
    log_level="INFO",
    keep_async_logs=False,
):
    """Download and extract ordinances for a list of jurisdictions

    This function scrapes ordinance documents (PDFs or HTML text) for a
    list of specified jurisdictions and processes them using one or more
    LLM models. Output files, logs, and intermediate artifacts are
    stored in configurable directories.

    Parameters
    ----------
    out_dir : path-like
        Path to the output directory. If it does not exist, it will be
        created. This directory will contain the structured ordinance
        CSV file, all downloaded ordinance documents (PDFs and HTML),
        usage metadata, and default subdirectories for logs and
        intermediate outputs (unless otherwise specified).
    tech : {"wind", "solar"}
        Label indicating which technology type is being processed.
    jurisdiction_fp : path-like
        Path to a CSV file specifying the jurisdictions to process.
        The CSV must contain at least two columns: "County" and "State",
        which specify the county and state names, respectively. If you
        would like to process a subdivision with a county, you must also
        include "Subdivision" and "Jurisdiction Type" columns. The
        "Subdivision" should be the name of the subdivision, and the
        "Jurisdiction Type" should be a string identifying the type of
        subdivision (e.g., "City", "Township", etc.)
    model : str or list of dict, optional
        LLM model(s) to use for scraping and parsing ordinance
        documents. If a string is provided, it is assumed to be the name
        of the default model (e.g., "gpt-4o"), and environment variables
        are used for authentication.

        If a list is provided, it should contain dictionaries of
        arguments that can initialize instances of
        :class:`~compass.llm.config.OpenAIConfig`. Each dictionary can
        specify the model name, client type, and initialization
        arguments.

        Each dictionary must also include a ``tasks`` key, which maps to
        a string or list of strings indicating the tasks that instance
        should handle. Exactly one of the instances **must** include
        "default" as a task, which will be used when no specific task is
        matched. For example::

            "model": [
                {
                    "model": "gpt-4o-mini",
                    "llm_call_kwargs": {
                        "temperature": 0,
                        "timeout": 300,
                    },
                    "client_kwargs": {
                        "api_key": "<your_api_key>",
                        "api_version": "<your_api_version>",
                        "azure_endpoint": "<your_azure_endpoint>",
                    },
                    "tasks": ["default", "date_extraction"],
                },
                {
                    "model": "gpt-4o",
                    "client_type": "openai",
                    "tasks": ["ordinance_text_extraction"],
                }
            ]

        By default, ``"gpt-4o"``.
    num_urls_to_check_per_jurisdiction : int, optional
        Number of unique Google search result URLs to check for each
        jurisdiction when attempting to locate ordinance documents.
        By default, ``5``.
    max_num_concurrent_browsers : int, optional
        Maximum number of browser instances to launch concurrently for
        retrieving information from the web. Increasing this value too
        much may lead to timeouts or performance issues on machines with
        limited resources. By default, ``10``.
    max_num_concurrent_website_searches : int, optional
        Maximum number of website searches allowed to run
        simultaneously. Increasing this value can speed up searches, but
        may lead to timeouts or performance issues on machines with
        limited resources. By default, ``10``.
    max_num_concurrent_jurisdictions : int, default=25
        Maximum number of jurisdictions to process concurrently.
        Limiting this can help manage memory usage when dealing with a
        large number of documents. By default ``25``.
    url_ignore_substrings : list of str, optional
        A list of substrings that, if found in any URL, will cause the
        URL to be excluded from consideration. This can be used to
        specify particular websites or entire domains to ignore. For
        example::

            url_ignore_substrings = [
                "wikipedia",
                "nrel.gov",
                "www.co.delaware.in.us/documents/1649699794_0382.pdf",
            ]

        The above configuration would ignore all `wikipedia` articles,
        all websites on the NREL domain, and the specific file located
        at `www.co.delaware.in.us/documents/1649699794_0382.pdf`.
        By default, ``None``.
    known_doc_urls : dict or str, optional
        A dictionary where keys are the jurisdiction codes (as strings)
        and values are a string or list of strings representing known
        URL's to check for those jurisdictions. If provided, these URL's
        will be checked first and if an ordinance document is found, no
        further scraping will be performed. This input can also be a
        path to a JSON file containing the dictionary of code-to-url
        mappings. By default, ``None``.
    file_loader_kwargs : dict, optional
        Dictionary of keyword arguments pairs to initialize
        :class:`elm.web.file_loader.AsyncFileLoader`. If found, the
        "pw_launch_kwargs" key in these will also be used to initialize
        the :class:`elm.web.search.google.PlaywrightGoogleLinkSearch`
        used for the google URL search. By default, ``None``.
    search_engines : list, optional
        A list of dictionaries, where each dictionary contains
        information about a search engine class that should be used for
        the document retrieval process. Each dictionary should contain
        at least the key ``"se_name"``, which should correspond to one
        of the search engine class names from
        :obj:`elm.web.search.run.SEARCH_ENGINE_OPTIONS`. The rest of the
        keys in the dictionary should contain keyword-value pairs to be
        used as parameters to initialize the search engine class (things
        like API keys and configuration options; see the ELM
        documentation for details on search engine class parameters).
        The list should be ordered by search engine preference - the
        first search engine parameters will be used to submit the
        queries initially, then any subsequent search engine listings
        will be used as fallback (in order that they appear). Do not
        repeat search engines - only the last config dictionary will be
        used to initialize the search engine if you do. If ``None``,
        then all default configurations for the search engines
        (along with the fallback order) are used. By default, ``None``.
    pytesseract_exe_fp : path-like, optional
        Path to the `pytesseract` executable. If specified, OCR will be
        used to extract text from scanned PDFs using Google's Tesseract.
        By default ``None``.
    td_kwargs : dict, optional
        Additional keyword arguments to pass to
        :class:`tempfile.TemporaryDirectory`. The temporary directory is
        used to store documents which have not yet been confirmed to
        contain relevant information. By default, ``None``.
    tpe_kwargs : dict, optional
        Additional keyword arguments to pass to
        :class:`concurrent.futures.ThreadPoolExecutor`, used for
        I/O-bound tasks such as logging. By default, ``None``.
    ppe_kwargs : dict, optional
        Additional keyword arguments to pass to
        :class:`concurrent.futures.ProcessPoolExecutor`, used for
        CPU-bound tasks such as PDF loading and parsing.
        By default, ``None``.
    log_dir : path-like, optional
        Path to the directory for storing log files. If not provided, a
        ``logs`` subdirectory will be created inside `out_dir`.
        By default, ``None``.
    clean_dir : path-like, optional
        Path to the directory for storing cleaned ordinance text output.
        If not provided, a ``cleaned_text`` subdirectory will be created
        inside `out_dir`. By default, ``None``.
    ordinance_file_dir : path-like, optional
        Path to the directory where downloaded ordinance files (PDFs or
        HTML) for each jurisdiction are stored. If not provided, a
        ``ordinance_files`` subdirectory will be created inside
        `out_dir`. By default, ``None``.
    jurisdiction_dbs_dir : path-like, optional
        Path to the directory where parsed ordinance database files are
        stored for each jurisdiction. If not provided, a
        ``jurisdiction_dbs`` subdirectory will be created inside
        `out_dir`. By default, ``None``.
    perform_website_search : bool, default=True
        Option to fallback to a jurisdiction website crawl-based search
        for ordinance documents if the search engine approach fails to
        recover any relevant documents. By default, ``True``.
    llm_costs : dict, optional
        Dictionary mapping model names to their token costs, used to
        track the estimated total cost of LLM usage during the run. The
        structure should be::

            {"model_name": {"prompt": float, "response": float}}

        Costs are specified in dollars per million tokens. For example::

            "llm_costs": {"my_gpt": {"prompt": 1.5, "response": 3.7}}

        registers a model named `"my_gpt"` with a cost of $1.5 per
        million input (prompt) tokens and $3.7 per million output
        (response) tokens for the current processing run.

        .. NOTE::

            The displayed total cost does not track cached tokens, so
            treat it like an estimate. Your final API costs may vary.

        If set to ``None``, no custom model costs are recorded, and
        cost tracking may be unavailable in the progress bar.
        By default, ``None``.
    log_level : str, optional
        Logging level for ordinance scraping and parsing (e.g., "TRACE",
        "DEBUG", "INFO", "WARNING", or "ERROR"). By default, ``"INFO"``.
    keep_async_logs : bool, default=False
        Option to store the full asynchronous log record to a file. This
        is only useful if you intend to monitor overall processing
        progress from a file instead of from the terminal. If ``True``,
        all of the unordered records are written to a "all.log" file in
        the `log_dir` directory. By default, ``False``.

    Returns
    -------
    str
        Message summarizing run results, including total processing
        time, total cost, output directory, and number of documents
        found. The message is formatted for easy reading in the terminal
        and may include color-coded cost information if the terminal
        supports it.
    """
    if log_level == "DEBUG":
        log_level = "DEBUG_TO_FILE"

    log_listener = LogListener(["compass", "elm"], level=log_level)
    LLM_COST_REGISTRY.update(llm_costs or {})
    dirs = _setup_folders(
        out_dir,
        log_dir=log_dir,
        clean_dir=clean_dir,
        ofd=ordinance_file_dir,
        jdd=jurisdiction_dbs_dir,
    )
    pk = ProcessKwargs(
        known_doc_urls,
        file_loader_kwargs,
        td_kwargs,
        tpe_kwargs,
        ppe_kwargs,
        max_num_concurrent_jurisdictions,
    )
    wsp = WebSearchParams(
        num_urls_to_check_per_jurisdiction,
        max_num_concurrent_browsers,
        max_num_concurrent_website_searches,
        url_ignore_substrings,
        pytesseract_exe_fp,
        search_engines,
    )
    models = _initialize_model_params(model)
    runner = _COMPASSRunner(
        dirs=dirs,
        log_listener=log_listener,
        tech=tech,
        models=models,
        web_search_params=wsp,
        process_kwargs=pk,
        perform_website_search=perform_website_search,
        log_level=log_level,
    )
    async with log_listener as ll:
        _setup_main_logging(dirs.logs, log_level, ll, keep_async_logs)
        return await runner.run(jurisdiction_fp)


class _COMPASSRunner:
    """Helper class to run COMPASS"""

    def __init__(
        self,
        dirs,
        log_listener,
        tech,
        models,
        web_search_params=None,
        process_kwargs=None,
        perform_website_search=True,
        log_level="INFO",
    ):
        self.dirs = dirs
        self.log_listener = log_listener
        self.tech = tech
        self.models = models
        self.web_search_params = web_search_params or WebSearchParams()
        self.process_kwargs = process_kwargs or ProcessKwargs()
        self.perform_website_search = perform_website_search
        self.log_level = log_level

    @cached_property
    def browser_semaphore(self):
        """asyncio.Semaphore or None: Sem to limit # of browsers"""
        return (
            asyncio.Semaphore(
                self.web_search_params.max_num_concurrent_browsers
            )
            if self.web_search_params.max_num_concurrent_browsers
            else None
        )

    @cached_property
    def crawl_semaphore(self):
        """asyncio.Semaphore or None: Sem to limit # of crawls"""
        return (
            asyncio.Semaphore(
                self.web_search_params.max_num_concurrent_website_searches
            )
            if self.web_search_params.max_num_concurrent_website_searches
            else None
        )

    @cached_property
    def search_engine_semaphore(self):
        """asyncio.Semaphore or None: Sem to limit # of SE queries"""
        return asyncio.Semaphore(MAX_CONCURRENT_SEARCH_ENGINE_QUERIES)

    @cached_property
    def _jurisdiction_semaphore(self):
        """asyncio.Semaphore or None: Sem to limit # of processes"""
        return (
            asyncio.Semaphore(
                self.process_kwargs.max_num_concurrent_jurisdictions
            )
            if self.process_kwargs.max_num_concurrent_jurisdictions
            else None
        )

    @property
    def jurisdiction_semaphore(self):
        """asyncio.Semaphore or AsyncExitStack: Jurisdictions limit"""
        if self._jurisdiction_semaphore is None:
            return AsyncExitStack()
        return self._jurisdiction_semaphore

    @cached_property
    def file_loader_kwargs(self):
        """dict: Keyword arguments for `AsyncFileLoader`"""
        file_loader_kwargs = _configure_file_loader_kwargs(
            self.process_kwargs.file_loader_kwargs
        )
        if self.web_search_params.pytesseract_exe_fp is not None:
            _setup_pytesseract(self.web_search_params.pytesseract_exe_fp)
            file_loader_kwargs.update(
                {"pdf_ocr_read_coroutine": read_pdf_doc_ocr}
            )
        return file_loader_kwargs

    @cached_property
    def known_doc_urls(self):
        """dict: Known URL's keyed by jurisdiction code"""
        known_doc_urls = self.process_kwargs.known_doc_urls or {}
        if isinstance(known_doc_urls, str):
            known_doc_urls = load_config(known_doc_urls)
        return {int(key): val for key, val in known_doc_urls.items()}

    @cached_property
    def tpe_kwargs(self):
        """dict: Keyword arguments for `ThreadPoolExecutor`"""
        return _configure_thread_pool_kwargs(self.process_kwargs.tpe_kwargs)

    @cached_property
    def _base_services(self):
        """list: List of required services to run for processing"""
        base_services = [
            TempFileCachePB(
                td_kwargs=self.process_kwargs.td_kwargs,
                tpe_kwargs=self.tpe_kwargs,
            ),
            TempFileCache(
                td_kwargs=self.process_kwargs.td_kwargs,
                tpe_kwargs=self.tpe_kwargs,
            ),
            FileMover(self.dirs.ordinance_files, tpe_kwargs=self.tpe_kwargs),
            CleanedFileWriter(
                self.dirs.clean_files, tpe_kwargs=self.tpe_kwargs
            ),
            OrdDBFileWriter(
                self.dirs.jurisdiction_dbs, tpe_kwargs=self.tpe_kwargs
            ),
            UsageUpdater(
                self.dirs.out / "usage.json", tpe_kwargs=self.tpe_kwargs
            ),
            JurisdictionUpdater(
                self.dirs.out / "jurisdictions.json",
                tpe_kwargs=self.tpe_kwargs,
            ),
            PDFLoader(**(self.process_kwargs.ppe_kwargs or {})),
        ]

        if self.web_search_params.pytesseract_exe_fp is not None:
            base_services.append(
                # pytesseract locks up with multiple processes, so
                # hardcode to only use 1 for now
                OCRPDFLoader(max_workers=1),
            )
        return base_services

    async def run(self, jurisdiction_fp):
        """Run COMPASS for a set of jurisdictions

        Parameters
        ----------
        jurisdiction_fp : path-like
            Path to CSV file containing the jurisdictions to search.

        Returns
        -------
        str
            Message summarizing run results, including total processing
            time, total cost, output directory, and number of documents
            found. The message is formatted for easy reading in the
            terminal and may include color-coded cost information if
            the terminal supports it.
        """
        logger.info("Running COMPASS version %s", __version__)
        jurisdictions = _load_jurisdictions_to_process(jurisdiction_fp)

        num_jurisdictions = len(jurisdictions)
        COMPASS_PB.create_main_task(num_jurisdictions=num_jurisdictions)
        start_date = datetime.now(UTC)

        doc_infos, total_cost = await self._run_all(jurisdictions)

        db, num_docs_found = doc_infos_to_db(doc_infos)
        save_db(db, self.dirs.out)
        total_time = save_run_meta(
            self.dirs,
            self.tech,
            start_date=start_date,
            end_date=datetime.now(UTC),
            num_jurisdictions_searched=num_jurisdictions,
            num_jurisdictions_found=num_docs_found,
            total_cost=total_cost,
            models=self.models,
        )
        run_msg = compile_run_summary_message(
            total_seconds=total_time,
            total_cost=total_cost,
            out_dir=self.dirs.out,
            document_count=num_docs_found,
        )
        for sub_msg in run_msg.split("\n"):
            logger.info(
                sub_msg.replace("[#71906e]", "").replace("[/#71906e]", "")
            )
        return run_msg

    async def _run_all(self, jurisdictions):
        """Process all jurisdictions with running services"""
        services = [model.llm_service for model in set(self.models.values())]
        services += self._base_services
        _ = self.file_loader_kwargs  # init loader kwargs once
        logger.info("Processing %d jurisdiction(s)", len(jurisdictions))
        async with RunningAsyncServices(services):
            tasks = []
            for __, row in jurisdictions.iterrows():
                jur_type, state, county, sub, fips, website = row[_JUR_COLS]
                jurisdiction = Jurisdiction(
                    subdivision_type=jur_type,
                    state=state,
                    county=county,
                    subdivision_name=sub,
                    code=fips,
                )
                usage_tracker = UsageTracker(
                    jurisdiction.full_name, usage_from_response
                )
                task = asyncio.create_task(
                    self._processed_jurisdiction_info_with_pb(
                        jurisdiction,
                        website,
                        self.known_doc_urls.get(fips),
                        usage_tracker=usage_tracker,
                    ),
                    name=jurisdiction.full_name,
                )
                tasks.append(task)
            doc_infos = await asyncio.gather(*tasks)
            total_cost = await _compute_total_cost()

        return doc_infos, total_cost

    async def _processed_jurisdiction_info_with_pb(
        self, jurisdiction, *args, **kwargs
    ):
        """Process jurisdiction and update progress bar"""
        async with self.jurisdiction_semaphore:
            with COMPASS_PB.jurisdiction_prog_bar(jurisdiction.full_name):
                return await self._processed_jurisdiction_info(
                    jurisdiction, *args, **kwargs
                )

    async def _processed_jurisdiction_info(self, *args, **kwargs):
        """Drop `doc` from RAM and only keep enough info to re-build"""

        doc = await self._process_jurisdiction_with_logging(*args, **kwargs)

        if doc is None or isinstance(doc, Exception):
            return None

        keys = ["source", "date", "jurisdiction", "ord_db_fp"]
        doc_info = {key: doc.attrs.get(key) for key in keys}
        logger.debug("Saving the following doc info:\n%s", doc_info)
        return doc_info

    async def _process_jurisdiction_with_logging(
        self,
        jurisdiction,
        jurisdiction_website,
        known_doc_urls=None,
        usage_tracker=None,
    ):
        """Retrieve ordinance document with async logs"""
        async with LocationFileLog(
            self.log_listener,
            self.dirs.logs,
            location=jurisdiction.full_name,
            level=self.log_level,
        ):
            task = asyncio.create_task(
                _SingleJurisdictionRunner(
                    self.tech,
                    jurisdiction,
                    self.models,
                    self.web_search_params,
                    self.file_loader_kwargs,
                    known_doc_urls=known_doc_urls,
                    browser_semaphore=self.browser_semaphore,
                    crawl_semaphore=self.crawl_semaphore,
                    search_engine_semaphore=self.search_engine_semaphore,
                    jurisdiction_website=jurisdiction_website,
                    perform_website_search=self.perform_website_search,
                    usage_tracker=usage_tracker,
                ).run(),
                name=jurisdiction.full_name,
            )
            try:
                doc, *__ = await asyncio.gather(task)
            except KeyboardInterrupt:
                raise
            except Exception as e:
                msg = "Encountered error of type %r while processing %s:"
                err_type = type(e)
                logger.exception(msg, err_type, jurisdiction.full_name)
                doc = None

            return doc


class _SingleJurisdictionRunner:
    """Helper class to process a single jurisdiction"""

    def __init__(  # noqa: PLR0913
        self,
        tech,
        jurisdiction,
        models,
        web_search_params,
        file_loader_kwargs,
        *,
        known_doc_urls=None,
        browser_semaphore=None,
        crawl_semaphore=None,
        search_engine_semaphore=None,
        jurisdiction_website=None,
        perform_website_search=True,
        usage_tracker=None,
    ):
        self.tech_specs = _compile_tech_specs(tech)
        self.jurisdiction = jurisdiction
        self.models = models
        self.web_search_params = web_search_params
        self.file_loader_kwargs = file_loader_kwargs
        self.known_doc_urls = known_doc_urls
        self.browser_semaphore = browser_semaphore
        self.crawl_semaphore = crawl_semaphore
        self.search_engine_semaphore = search_engine_semaphore
        self.usage_tracker = usage_tracker
        self.jurisdiction_website = jurisdiction_website
        self.perform_website_search = perform_website_search
        self.validate_user_website_input = True
        self._jsp = None

    @cached_property
    def file_loader_kwargs_no_ocr(self):
        """dict: Keyword arguments for `AsyncFileLoader` with no OCR"""
        flk = deepcopy(self.file_loader_kwargs)
        flk.pop("pdf_ocr_read_coroutine", None)
        return flk

    @contextmanager
    def _tracked_progress(self):
        """Context manager to set up jurisdiction sub-progress bar"""
        loc = self.jurisdiction.full_name
        with COMPASS_PB.jurisdiction_sub_prog(loc) as self._jsp:
            yield

        self._jsp = None

    async def run(self):
        """Download and parse document for a single jurisdiction"""
        start_time = time.monotonic()
        doc = None
        try:
            doc = await self._run()
        finally:
            await self._record_usage()
            await _record_jurisdiction_info(
                self.jurisdiction, doc, start_time, self.usage_tracker
            )

        return doc

    async def _run(self):
        """Search for docs and parse them for ordinances"""
        if self.known_doc_urls:
            docs = await self._download_known_url_documents()
            if docs is not None:
                COMPASS_PB.update_jurisdiction_task(
                    self.jurisdiction.full_name,
                    description="Extracting structured data...",
                )
                doc = await self._parse_docs_for_ordinances(docs)
            else:
                doc = None

            if doc is not None:
                return doc

        docs = await self._find_documents_using_search_engine()
        if docs is not None:
            COMPASS_PB.update_jurisdiction_task(
                self.jurisdiction.full_name,
                description="Extracting structured data...",
            )
            doc = await self._parse_docs_for_ordinances(docs)
        else:
            doc = None

        if doc is not None or not self.perform_website_search:
            return doc

        docs = await self._find_documents_from_website()
        if not docs:
            return None

        COMPASS_PB.update_jurisdiction_task(
            self.jurisdiction.full_name,
            description="Extracting structured data...",
        )
        return await self._parse_docs_for_ordinances(docs)

    async def _download_known_url_documents(self):
        """Download ordinance documents from known URLs"""

        if isinstance(self.known_doc_urls, str):
            self.known_doc_urls = [self.known_doc_urls]

        docs = await download_known_urls(
            self.jurisdiction,
            self.known_doc_urls,
            browser_semaphore=self.browser_semaphore,
            file_loader_kwargs=self.file_loader_kwargs,
        )

        if not docs:
            return None

        docs = await filter_ordinance_docs(
            docs,
            self.jurisdiction,
            self.models,
            heuristic=self.tech_specs.heuristic,
            tech=self.tech_specs.name,
            ordinance_text_collector_class=(
                self.tech_specs.ordinance_text_collector
            ),
            permitted_use_text_collector_class=(
                self.tech_specs.permitted_use_text_collector
            ),
            usage_tracker=self.usage_tracker,
            check_for_correct_jurisdiction=False,
        )
        if not docs:
            return None

        for doc in docs:
            doc.attrs["jurisdiction"] = self.jurisdiction
            doc.attrs["jurisdiction_name"] = self.jurisdiction.full_name
            doc.attrs["jurisdiction_website"] = None
            doc.attrs["compass_crawl"] = False

        await self._record_usage()
        return docs

    async def _find_documents_using_search_engine(self):
        """Search the web for an ordinance document and construct it"""
        docs = await download_jurisdiction_ordinance_using_search_engine(
            self.tech_specs.questions,
            self.jurisdiction,
            num_urls=self.web_search_params.num_urls_to_check_per_jurisdiction,
            file_loader_kwargs=self.file_loader_kwargs,
            search_semaphore=self.search_engine_semaphore,
            browser_semaphore=self.browser_semaphore,
            url_ignore_substrings=self.web_search_params.url_ignore_substrings,
            **self.web_search_params.se_kwargs,
        )
        docs = await filter_ordinance_docs(
            docs,
            self.jurisdiction,
            self.models,
            heuristic=self.tech_specs.heuristic,
            tech=self.tech_specs.name,
            ordinance_text_collector_class=(
                self.tech_specs.ordinance_text_collector
            ),
            permitted_use_text_collector_class=(
                self.tech_specs.permitted_use_text_collector
            ),
            usage_tracker=self.usage_tracker,
            check_for_correct_jurisdiction=True,
        )
        if not docs:
            return None

        for doc in docs:
            doc.attrs["jurisdiction"] = self.jurisdiction
            doc.attrs["jurisdiction_name"] = self.jurisdiction.full_name
            doc.attrs["jurisdiction_website"] = None
            doc.attrs["compass_crawl"] = False

        await self._record_usage()
        return docs

    async def _find_documents_from_website(self):
        """Search the website for ordinance documents"""
        if self.jurisdiction_website and self.validate_user_website_input:
            await self._validate_jurisdiction_website()

        if not self.jurisdiction_website:
            website = await self._try_find_jurisdiction_website()
            if not website:
                return None
            self.jurisdiction_website = website

        docs, scrape_results = await self._try_elm_crawl()

        found_with_compass_crawl = False
        if not docs:
            docs = await self._try_compass_crawl(scrape_results)
            found_with_compass_crawl = True

        if not docs:
            return None

        for doc in docs:
            doc.attrs["jurisdiction"] = self.jurisdiction
            doc.attrs["jurisdiction_name"] = self.jurisdiction.full_name
            doc.attrs["jurisdiction_website"] = self.jurisdiction_website
            doc.attrs["compass_crawl"] = found_with_compass_crawl

        await self._record_usage()
        return docs

    async def _validate_jurisdiction_website(self):
        """Validate user input for jurisdiction website"""
        if self.jurisdiction_website is None:
            return

        self.jurisdiction_website = await get_redirected_url(
            self.jurisdiction_website, timeout=30
        )
        COMPASS_PB.update_jurisdiction_task(
            self.jurisdiction.full_name,
            description=(
                f"Validating user input website: {self.jurisdiction_website}"
            ),
        )
        model_config = self.models.get(
            LLMTasks.DOCUMENT_JURISDICTION_VALIDATION,
            self.models[LLMTasks.DEFAULT],
        )
        validator = JurisdictionWebsiteValidator(
            browser_semaphore=self.browser_semaphore,
            file_loader_kwargs=self.file_loader_kwargs_no_ocr,
            usage_tracker=self.usage_tracker,
            llm_service=model_config.llm_service,
            **model_config.llm_call_kwargs,
        )
        is_website_correct = await validator.check(
            self.jurisdiction_website, self.jurisdiction
        )
        if not is_website_correct:
            self.jurisdiction_website = None

    async def _try_find_jurisdiction_website(self):
        """Use web to try to find the main jurisdiction website"""
        COMPASS_PB.update_jurisdiction_task(
            self.jurisdiction.full_name,
            description="Searching for jurisdiction website...",
        )
        return await find_jurisdiction_website(
            self.jurisdiction,
            self.models,
            file_loader_kwargs=self.file_loader_kwargs_no_ocr,
            search_semaphore=self.search_engine_semaphore,
            browser_semaphore=self.browser_semaphore,
            usage_tracker=self.usage_tracker,
            url_ignore_substrings=(
                self.web_search_params.url_ignore_substrings
            ),
            **self.web_search_params.se_kwargs,
        )

    async def _try_elm_crawl(self):
        """Try crawling website using ELM crawler"""
        self.jurisdiction_website = await get_redirected_url(
            self.jurisdiction_website, timeout=30
        )
        out = await download_jurisdiction_ordinances_from_website(
            self.jurisdiction_website,
            heuristic=self.tech_specs.heuristic,
            keyword_points=self.tech_specs.website_url_keyword_points,
            file_loader_kwargs=self.file_loader_kwargs_no_ocr,
            crawl_semaphore=self.crawl_semaphore,
            pb_jurisdiction_name=self.jurisdiction.full_name,
            return_c4ai_results=True,
        )
        docs, scrape_results = out
        docs = await filter_ordinance_docs(
            docs,
            self.jurisdiction,
            self.models,
            heuristic=self.tech_specs.heuristic,
            tech=self.tech_specs.name,
            ordinance_text_collector_class=(
                self.tech_specs.ordinance_text_collector
            ),
            permitted_use_text_collector_class=(
                self.tech_specs.permitted_use_text_collector
            ),
            usage_tracker=self.usage_tracker,
            check_for_correct_jurisdiction=True,
        )
        return docs, scrape_results

    async def _try_compass_crawl(self, scrape_results):
        """Try to crawl the website with compass-style crawling"""
        checked_urls = set()
        for scrape_result in scrape_results:
            checked_urls.update({sub_res.url for sub_res in scrape_result})
        docs = (
            await download_jurisdiction_ordinances_from_website_compass_crawl(
                self.jurisdiction_website,
                heuristic=self.tech_specs.heuristic,
                keyword_points=self.tech_specs.website_url_keyword_points,
                file_loader_kwargs=self.file_loader_kwargs_no_ocr,
                already_visited=checked_urls,
                crawl_semaphore=self.crawl_semaphore,
                pb_jurisdiction_name=self.jurisdiction.full_name,
            )
        )
        return await filter_ordinance_docs(
            docs,
            self.jurisdiction,
            self.models,
            heuristic=self.tech_specs.heuristic,
            tech=self.tech_specs.name,
            ordinance_text_collector_class=(
                self.tech_specs.ordinance_text_collector
            ),
            permitted_use_text_collector_class=(
                self.tech_specs.permitted_use_text_collector
            ),
            usage_tracker=self.usage_tracker,
            check_for_correct_jurisdiction=True,
        )

    async def _parse_docs_for_ordinances(self, docs):
        """Parse docs (in order) for ordinances"""
        for possible_ord_doc in docs:
            doc = await self._try_extract_all_ordinances(possible_ord_doc)
            ord_count = num_ordinances_in_doc(
                doc, exclude_features=EXCLUDE_FROM_ORD_DOC_CHECK
            )
            if ord_count > 0:
                logger.debug(
                    "Found ordinances in doc from %s",
                    possible_ord_doc.attrs.get("source", "unknown source"),
                )
                return await _move_files(doc, self.jurisdiction)

        logger.debug("No ordinances found; searched %d docs", len(docs))
        return None

    async def _try_extract_all_ordinances(self, possible_ord_doc):
        """Try to extract ordinance values and permitted districts"""
        with self._tracked_progress():
            tasks = [
                asyncio.create_task(
                    self._try_extract_ordinances(possible_ord_doc, **kwargs),
                    name=self.jurisdiction.full_name,
                )
                for kwargs in self._extraction_task_kwargs
            ]

            docs = await asyncio.gather(*tasks)

        return _concat_scrape_results(docs[0])

    @property
    def _extraction_task_kwargs(self):
        """Keyword-argument pairs to pass to _try_extract_ordinances"""
        return [
            {
                "extractor_class": self.tech_specs.ordinance_text_extractor,
                "original_text_key": "ordinance_text",
                "cleaned_text_key": "cleaned_ordinance_text",
                "text_model": self.models.get(
                    LLMTasks.ORDINANCE_TEXT_EXTRACTION,
                    self.models[LLMTasks.DEFAULT],
                ),
                "parser_class": self.tech_specs.structured_ordinance_parser,
                "out_key": "ordinance_values",
                "value_model": self.models.get(
                    LLMTasks.ORDINANCE_VALUE_EXTRACTION,
                    self.models[LLMTasks.DEFAULT],
                ),
            },
            {
                "extractor_class": (
                    self.tech_specs.permitted_use_text_extractor
                ),
                "original_text_key": "permitted_use_text",
                "cleaned_text_key": "districts_text",
                "text_model": self.models.get(
                    LLMTasks.PERMITTED_USE_TEXT_EXTRACTION,
                    self.models[LLMTasks.DEFAULT],
                ),
                "parser_class": (
                    self.tech_specs.structured_permitted_use_parser
                ),
                "out_key": "permitted_district_values",
                "value_model": self.models.get(
                    LLMTasks.PERMITTED_USE_VALUE_EXTRACTION,
                    self.models[LLMTasks.DEFAULT],
                ),
            },
        ]

    async def _try_extract_ordinances(
        self,
        possible_ord_doc,
        extractor_class,
        original_text_key,
        cleaned_text_key,
        parser_class,
        out_key,
        text_model,
        value_model,
    ):
        """Try applying a single extractor to the relevant legal text"""
        logger.debug(
            "Checking for ordinances in doc from %s",
            possible_ord_doc.attrs.get("source", "unknown source"),
        )
        assert self._jsp is not None, "No progress bar set!"
        task_id = self._jsp.add_task(_TEXT_EXTRACTION_TASKS[extractor_class])
        doc = await _extract_ordinance_text(
            possible_ord_doc,
            extractor_class=extractor_class,
            original_text_key=original_text_key,
            usage_tracker=self.usage_tracker,
            model_config=text_model,
        )
        await self._record_usage()
        self._jsp.remove_task(task_id)
        out = await _extract_ordinances_from_text(
            doc,
            parser_class=parser_class,
            text_key=cleaned_text_key,
            out_key=out_key,
            usage_tracker=self.usage_tracker,
            model_config=value_model,
        )
        await self._record_usage()
        return out

    async def _record_usage(self):
        """Dump usage to file if tracker given"""
        if self.usage_tracker is None:
            return

        total_usage = await UsageUpdater.call(self.usage_tracker)
        total_cost = _compute_total_cost_from_usage(total_usage)
        COMPASS_PB.update_total_cost(total_cost, replace=True)


def _compile_tech_specs(tech):
    """Compile `TechSpec` tuple based on the user `tech` input"""
    if tech.casefold() == "wind":
        return TechSpec(
            "wind",
            WIND_QUESTION_TEMPLATES,
            WindHeuristic(),
            WindOrdinanceTextCollector,
            WindOrdinanceTextExtractor,
            WindPermittedUseDistrictsTextCollector,
            WindPermittedUseDistrictsTextExtractor,
            StructuredWindOrdinanceParser,
            StructuredWindPermittedUseDistrictsParser,
            BEST_WIND_ORDINANCE_WEBSITE_URL_KEYWORDS,
        )
    if tech.casefold() == "solar":
        return TechSpec(
            "solar",
            SOLAR_QUESTION_TEMPLATES,
            SolarHeuristic(),
            SolarOrdinanceTextCollector,
            SolarOrdinanceTextExtractor,
            SolarPermittedUseDistrictsTextCollector,
            SolarPermittedUseDistrictsTextExtractor,
            StructuredSolarOrdinanceParser,
            StructuredSolarPermittedUseDistrictsParser,
            BEST_SOLAR_ORDINANCE_WEBSITE_URL_KEYWORDS,
        )

    msg = f"Unknown tech input: {tech}"
    raise COMPASSValueError(msg)


def _setup_main_logging(log_dir, level, listener, keep_async_logs):
    """Setup main logger for catching exceptions during execution"""
    fmt = logging.Formatter(fmt="[%(asctime)s] %(levelname)s: %(message)s")
    handler = logging.FileHandler(log_dir / "main.log", encoding="utf-8")
    handler.setFormatter(fmt)
    handler.setLevel(level)
    handler.addFilter(NoLocationFilter())
    listener.addHandler(handler)

    if keep_async_logs:
        handler = logging.FileHandler(log_dir / "all.log", encoding="utf-8")
        fmt = logging.Formatter(
            fmt="[%(asctime)s] %(levelname)s - %(taskName)s: %(message)s",
        )
        handler.setFormatter(fmt)
        handler.setLevel(level)
        listener.addHandler(handler)


def _setup_folders(out_dir, log_dir=None, clean_dir=None, ofd=None, jdd=None):
    """Setup output directory folders"""
    dirs = Directories(out_dir, log_dir, clean_dir, ofd, jdd)

    if dirs.out.exists():
        msg = (
            f"Output directory '{out_dir!s}' already exists! Please specify a "
            "new directory for every COMPASS run."
        )
        raise COMPASSValueError(msg)

    dirs.make_dirs()
    return dirs


def _initialize_model_params(user_input):
    """Initialize llm caller args for models from user input"""
    if isinstance(user_input, str):
        return {LLMTasks.DEFAULT: OpenAIConfig(name=user_input)}

    caller_instances = {}
    for kwargs in user_input:
        tasks = kwargs.pop("tasks", LLMTasks.DEFAULT)
        if isinstance(tasks, str):
            tasks = [tasks]

        model_config = OpenAIConfig(**kwargs)
        for task in tasks:
            if task in caller_instances:
                msg = (
                    f"Found duplicated task: {task!r}. Please ensure each "
                    "LLM caller definition has uniquely-assigned tasks."
                )
                raise COMPASSValueError(msg)
            caller_instances[task] = model_config

    return caller_instances


def _load_jurisdictions_to_process(jurisdiction_fp):
    """Load the jurisdictions to retrieve documents for"""
    if jurisdiction_fp is None:
        logger.info("No `jurisdiction_fp` input! Loading all jurisdictions")
        return load_all_jurisdiction_info()
    return load_jurisdictions_from_fp(jurisdiction_fp)


def _configure_thread_pool_kwargs(tpe_kwargs):
    """Set thread pool workers to 5 if user didn't specify"""
    tpe_kwargs = tpe_kwargs or {}
    tpe_kwargs.setdefault("max_workers", 5)
    return tpe_kwargs


def _configure_file_loader_kwargs(file_loader_kwargs):
    """Add PDF reading coroutine to kwargs"""
    file_loader_kwargs = file_loader_kwargs or {}
    file_loader_kwargs.update({"pdf_read_coroutine": read_pdf_doc})
    return file_loader_kwargs


async def _extract_ordinance_text(
    doc, extractor_class, original_text_key, usage_tracker, model_config
):
    """Extract text pertaining to ordinance of interest"""
    llm_caller = LLMCaller(
        llm_service=model_config.llm_service,
        usage_tracker=usage_tracker,
        **model_config.llm_call_kwargs,
    )
    extractor = extractor_class(llm_caller)
    doc = await extract_ordinance_text_with_ngram_validation(
        doc,
        model_config.text_splitter,
        extractor,
        original_text_key=original_text_key,
    )
    return await _write_cleaned_text(doc)


async def _extract_ordinances_from_text(
    doc, parser_class, text_key, out_key, usage_tracker, model_config
):
    """Extract values from ordinance text"""
    parser = parser_class(
        llm_service=model_config.llm_service,
        usage_tracker=usage_tracker,
        **model_config.llm_call_kwargs,
    )
    logger.info("Extracting %s...", out_key.replace("_", " "))
    return await extract_ordinance_values(
        doc, parser, text_key=text_key, out_key=out_key
    )


async def _move_files(doc, jurisdiction):
    """Move files to output folders, if applicable"""
    ord_count = num_ordinances_in_doc(doc)
    if ord_count == 0:
        logger.info("No ordinances found for %s.", jurisdiction.full_name)
        return doc

    doc = await _move_file_to_out_dir(doc)
    doc = await _write_ord_db(doc)
    logger.info(
        "%d ordinance value(s) found for %s. Outputs are here: '%s'",
        ord_count,
        jurisdiction.full_name,
        doc.attrs["ord_db_fp"],
    )
    return doc


async def _move_file_to_out_dir(doc):
    """Move PDF or HTML text file to output directory"""
    out_fp = await FileMover.call(doc)
    doc.attrs["out_fp"] = out_fp
    return doc


async def _write_cleaned_text(doc):
    """Write cleaned text to `clean_files` dir"""
    out_fp = await CleanedFileWriter.call(doc)
    doc.attrs["cleaned_fps"] = out_fp
    return doc


async def _write_ord_db(doc):
    """Write cleaned text to `jurisdiction_dbs` dir"""
    out_fp = await OrdDBFileWriter.call(doc)
    doc.attrs["ord_db_fp"] = out_fp
    return doc


async def _record_jurisdiction_info(loc, doc, start_time, usage_tracker):
    """Record info about jurisdiction"""
    seconds_elapsed = time.monotonic() - start_time
    await JurisdictionUpdater.call(loc, doc, seconds_elapsed, usage_tracker)


def _setup_pytesseract(exe_fp):
    """Set the pytesseract command"""
    import pytesseract  # noqa: PLC0415

    logger.debug("Setting `tesseract_cmd` to %s", exe_fp)
    pytesseract.pytesseract.tesseract_cmd = exe_fp


def _concat_scrape_results(doc):
    data = [
        doc.attrs.get(key, None)
        for key in ["ordinance_values", "permitted_district_values"]
    ]
    data = [df for df in data if df is not None and not df.empty]
    if len(data) == 0:
        return doc

    if len(data) == 1:
        doc.attrs["scraped_values"] = data[0]
        return doc

    doc.attrs["scraped_values"] = pd.concat(data)
    return doc


async def _compute_total_cost():
    """Compute total cost from tracked usage"""
    total_usage = await UsageUpdater.call(None)
    if not total_usage:
        return 0

    return _compute_total_cost_from_usage(total_usage)


def _compute_total_cost_from_usage(tracked_usage):
    """Compute total cost from total tracked usage"""

    total_cost = 0
    for usage in tracked_usage.values():
        totals = usage.get("tracker_totals", {})
        for model, total_usage in totals.items():
            model_costs = LLM_COST_REGISTRY.get(model, {})
            total_cost += (
                total_usage.get("prompt_tokens", 0)
                / 1e6
                * model_costs.get("prompt", 0)
            )
            total_cost += (
                total_usage.get("response_tokens", 0)
                / 1e6
                * model_costs.get("response", 0)
            )

    return total_cost



"""COMPASS utilities for finalizing a run directory"""

import json
import getpass
import logging
from pathlib import Path

import pandas as pd
from elm.version import __version__ as elm_version

from compass import __version__ as compass_version
from compass.utilities.parsing import (
    extract_ord_year_from_doc_attrs,
    num_ordinances_dataframe,
    ordinances_bool_index,
)


logger = logging.getLogger(__name__)
_PARSED_COLS = [
    # TODO: Put these in an enum
    "county",
    "state",
    "subdivision",
    "jurisdiction_type",
    "FIPS",
    "feature",
    "value",
    "units",
    "adder",
    "min_dist",
    "max_dist",
    "summary",
    "ord_year",
    "section",
    "source",
    "quantitative",
]
QUANT_OUT_COLS = _PARSED_COLS[:-1]
"""Output columns in quantitative ordinance file"""
QUAL_OUT_COLS = _PARSED_COLS[:6] + _PARSED_COLS[-5:-1]
"""Output columns in qualitative ordinance file"""


def save_run_meta(
    dirs,
    tech,
    start_date,
    end_date,
    num_jurisdictions_searched,
    num_jurisdictions_found,
    total_cost,
    models,
):
    """Write out meta information about ordinance collection run

    Parameters
    ----------
    dirs : :class:`~compass.utilities.base.Directories`
        Directories instance containing information about the output
        directories used for the run.
    tech : {"wind", "solar"}
        Technology that was the target of the run.
    start_date, end_date : datetime.datetime
        Instances representing the start and end dates, respectively.
    num_jurisdictions_searched, num_jurisdictions_found : int
        Total number of jurisdictions that were searched and actually
        found, respectively.
    total_cost : float
        Total cost of the processing, in $.
    models : dict
        Dictionary mapping task names (from
        :class:`~compass.utilities.enums.LLMTasks`) to
        :class:`~compass.llm.config.OpenAIConfig` instances used for the
        run.

    Returns
    -------
    run_time : float
        Total processing run-time, in seconds.
    """

    try:
        username = getpass.getuser()
    except OSError:
        username = "Unknown"

    time_elapsed = end_date - start_date
    meta_data = {
        "username": username,
        "versions": {"elm": elm_version, "compass": compass_version},
        "technology": tech,
        "models": _extract_model_info_from_all_models(models),
        "time_start_utc": start_date.isoformat(),
        "time_end_utc": end_date.isoformat(),
        "total_time": time_elapsed.seconds,
        "total_time_string": str(time_elapsed),
        "num_jurisdictions_searched": num_jurisdictions_searched,
        "num_jurisdictions_found": num_jurisdictions_found,
        "cost": total_cost or None,
        "manifest": {},
    }
    manifest = {
        "LOG_DIR": dirs.logs,
        "CLEAN_FILE_DIR": dirs.clean_files,
        "JURISDICTION_DBS_DIR": dirs.jurisdiction_dbs,
        "ORDINANCE_FILES_DIR": dirs.ordinance_files,
        "USAGE_FILE": dirs.out / "usage.json",
        "JURISDICTION_FILE": dirs.out / "jurisdictions.json",
        "QUANT_DATA_FILE": dirs.out / "quantitative_ordinances.csv",
        "QUAL_DATA_FILE": dirs.out / "quantitative_ordinances.csv",
    }
    for name, file_path in manifest.items():
        if file_path.exists():
            meta_data["manifest"][name] = str(file_path.relative_to(dirs.out))
        else:
            meta_data["manifest"][name] = None

    meta_data["manifest"]["META_FILE"] = "meta.json"
    with (dirs.out / "meta.json").open("w", encoding="utf-8") as fh:
        json.dump(meta_data, fh, indent=4)

    return time_elapsed.seconds


def doc_infos_to_db(doc_infos):
    """Convert list of docs to output database

    Parameters
    ----------
    doc_infos : iterable of dict
        Iterable of dictionaries, where each dictionary has at least the
        following keys:

            - "ord_db_fp": Path to parsed ordinance CSV file
            - "source": URL of the file from which ordinances were
              extracted
            - "date": Tuple of (year, month, day). Any of the values can
              be ``None``.
            - "jurisdiction": Instance of Jurisdiction representing the
              jurisdiction associated with these ordinance values.

        If this iterable is empty, and empty DataFrame (with the correct
        columns) is returned.

    Returns
    -------
    ordinances : pandas.DataFrame
        DataFrame containing ordinances collected from all individual
        CSV's.
    count : int
        Total number jurisdictions for which ordinances were found.
    """
    db = []
    for doc_info in doc_infos:
        if doc_info is None:
            continue

        ord_db_fp = doc_info.get("ord_db_fp")
        if ord_db_fp is None:
            continue

        ord_db = pd.read_csv(ord_db_fp)

        if num_ordinances_dataframe(ord_db) == 0:
            continue

        results = _db_results(ord_db, doc_info)
        results = _formatted_db(results)
        db.append(results)

    if not db:
        return pd.DataFrame(columns=_PARSED_COLS), 0

    logger.info("Compiling final database for %d jurisdiction(s)", len(db))
    num_jurisdictions_found = len(db)
    db = pd.concat([df.dropna(axis=1, how="all") for df in db], axis=0)
    db = _empirical_adjustments(db)
    return _formatted_db(db), num_jurisdictions_found


def save_db(db, out_dir):
    """Split DB into qualitative vs quantitative and save to disk

    Parameters
    ----------
    db : pandas.DataFrame
        Pandas DataFrame containing ordinance data to save. Must have
        all columns in :obj:`QUANT_OUT_COLS` and :obj:`QUAL_OUT_COLS`
        as well as a ``"quantitative"`` column that contains a boolean
        determining whether the rwo belongs in the quantitative output
        file (``True``) or the qualitative output file (``False``).
    out_dir : path-like
        Path to output directory where ordinance database csv files
        should be written.
    """
    if db.empty:
        return

    out_dir = Path(out_dir)
    qual_db = db[~db["quantitative"]][QUAL_OUT_COLS]
    quant_db = db[db["quantitative"]][QUANT_OUT_COLS]
    qual_db.to_csv(out_dir / "qualitative_ordinances.csv", index=False)
    quant_db.to_csv(out_dir / "quantitative_ordinances.csv", index=False)


def _db_results(results, doc_info):
    """Extract results from doc attrs to DataFrame"""

    results["source"] = doc_info.get("source")
    results["ord_year"] = extract_ord_year_from_doc_attrs(doc_info)

    jurisdiction = doc_info["jurisdiction"]
    results["FIPS"] = jurisdiction.code
    results["county"] = jurisdiction.county
    results["state"] = jurisdiction.state
    results["subdivision"] = jurisdiction.subdivision_name
    results["jurisdiction_type"] = jurisdiction.type
    return results


def _empirical_adjustments(db):
    """Post-processing adjustments based on empirical observations

    Current adjustments include:

        - Limit adder to max of 250 ft.
            - Chat GPT likes to report large values here, but in
            practice all values manually observed in ordinance documents
            are below 250 ft. If large value is detected, assume it's an
            error on Chat GPT's part and remove it.

    """
    if "adder" in db.columns:
        db.loc[db["adder"] > 250, "adder"] = None  # noqa: PLR2004
    return db


def _formatted_db(db):
    """Format DataFrame for output"""
    for col in _PARSED_COLS:
        if col not in db.columns:
            db[col] = None

    db["quantitative"] = db["quantitative"].astype("boolean").fillna(True)
    ord_rows = ordinances_bool_index(db)
    return db[ord_rows][_PARSED_COLS].reset_index(drop=True)


def _extract_model_info_from_all_models(models):
    """Group model info together"""
    models_to_tasks = {}
    for task, caller_args in models.items():
        models_to_tasks.setdefault(caller_args, []).append(task)

    return [
        {
            "name": caller_args.name,
            "llm_call_kwargs": caller_args.llm_call_kwargs or None,
            "llm_service_rate_limit": caller_args.llm_service_rate_limit,
            "text_splitter_chunk_size": caller_args.text_splitter_chunk_size,
            "text_splitter_chunk_overlap": (
                caller_args.text_splitter_chunk_overlap
            ),
            "client_type": caller_args.client_type,
            "tasks": tasks,
        }
        for caller_args, tasks in models_to_tasks.items()
    ]


def compile_run_summary_message(
    total_seconds, total_cost, out_dir, document_count
):
    """Summarize the run results into a formatted string

    Parameters
    ----------
    total_seconds : int or float
        Total number of seconds the run took to complete.
    total_cost : int or float
        Total cost of the run, in $.
    out_dir : path-like
        Path to output directory where the run results are saved.
    document_count : int
        Number of documents found during the run.

    Returns
    -------
    str
        Formatted string summarizing the run results.
    """
    runtime = _elapsed_time_as_str(total_seconds)
    total_cost = (
        f"\nTotal cost: [#71906e]${total_cost:,.2f}[/#71906e]"
        if total_cost
        else ""
    )

    return (
        f"✅ Scraping complete!\nOutput Directory: {out_dir}\n"
        f"Total runtime: {runtime} {total_cost}\n"
        f"Number of documents found: {document_count}"
    )


def _elapsed_time_as_str(seconds_elapsed):
    """Format elapsed time into human readable string"""
    days, seconds = divmod(int(seconds_elapsed), 24 * 3600)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    time_str = f"{hours:d}:{minutes:02d}:{seconds:02d}"
    if days:
        time_str = f"{days:,d} day{'s' if abs(days) != 1 else ''}, {time_str}"
    return time_str



"""Custom COMPASS website crawler

Much more simplistic than the Crawl4AI crawler, but designed to access
some links that Crawl4AI cannot (such as those behind a button
interface).
"""

import logging
import operator
from collections import Counter
from contextlib import AsyncExitStack
from urllib.parse import (
    urlparse,
    urlunparse,
    quote,
    unquote,
    parse_qsl,
    urlencode,
    urljoin,
)

from crawl4ai.models import Link as c4AILink
from bs4 import BeautifulSoup
from rebrowser_playwright.async_api import async_playwright
from rebrowser_playwright.async_api import Error as RBPlaywrightError
from playwright._impl._errors import Error as PlaywrightError  # noqa: PLC2701
from elm.web.utilities import pw_page
from elm.web.document import PDFDocument, HTMLDocument
from elm.web.file_loader import AsyncFileLoader
from elm.web.website_crawl import ELMLinkScorer, _SCORE_KEY  # noqa: PLC2701


logger = logging.getLogger(__name__)
_DEPTH_KEY = "web_crawl_depth"
_CLICKABLE_SELECTORS = [
    "button",  # "a", "p"
]
_BLACKLIST_SUBSTRINGS = [
    "facebook",
    "twitter",
    "linkedin",
    "instagram",
    "youtube",
    "instagram",
    "pinterest",  # cspell: disable-line
    "tiktok",  # cspell: disable-line
    "x.com",
    "snapchat",
    "reddit",
    "mailto:",
    "tel:",
    "javascript:",
    "login",
    "signup",
    "sign up",
    "signin",
    "sign in",
    "register",
    "subscribe",
    "donate",
    "shop",
    "cart",
    "careers",
    "event",
    "events",
    "calendar",
]
DOC_THRESHOLD = 5
"""Default max documents to collect before terminating COMPASS crawl"""


class COMPASSLinkScorer(ELMLinkScorer):
    """Custom URL scorer for COMPASS website crawling"""

    def _assign_value(self, text):
        """Score based on the presence of keywords in link text"""
        score = 0
        text = text.casefold().replace("plant", "")
        for kw, kw_score in self.keyword_points.items():
            if kw in text:
                score += kw_score
        return score


class _Link(c4AILink):
    """Crawl4AI Link subclass with a few utilities"""

    def __hash__(self):
        return hash(self.href.casefold())

    def __repr__(self):
        return (
            f"Link(title={self.title!r}, href={self.href!r}, "
            f"base_domain={self.base_domain!r})"
        )

    def __str__(self):
        return f"{self.title} ({self.href})"

    def __eq__(self, other):
        if isinstance(other, str):
            return self.href.casefold() == other.casefold()

        if not isinstance(other, c4AILink):
            return NotImplemented
        return self.href.casefold() == other.href.casefold()

    @property
    def consistent_domain(self):
        """bool: ``True`` if the link is from the base domain"""
        return self.base_domain.casefold() in self.href.casefold()

    @property
    def resembles_pdf(self):
        """bool: ``True`` if the link has "pdf" in title or href"""
        return "pdf" in self.title.casefold() or "pdf" in self.href.casefold()


class COMPASSCrawler:
    """A simple website crawler to search for ordinance documents"""

    def __init__(
        self,
        validator,
        url_scorer,
        file_loader_kwargs=None,
        already_visited=None,
        num_link_scores_to_check_per_page=4,
        max_pages=100,
        browser_semaphore=None,
    ):
        """

        Parameters
        ----------
        validator : callable
            An async callable that takes a document instance (containing
            the text from a PDF or a webpage) and returns a boolean
            indicating whether the text passes the validation check.
            This is used to determine whether or not to keep (i.e.
            return) the document.
        url_scorer : callable
            An async callable that takes a list of dictionaries
            containing URL information and assigns each dictionary a
            `score` key representing the score for that URL. The input
            URL dictionaries will each have at least one key: "href".
            This key will contain the URL of the link. The dictionary
            may also have other attributes such as "title", which
            contains the link title text.
        file_loader_kwargs : dict, optional
            Additional keyword-value argument pairs to pass to the
            :class:`~elm.web.file_loader.AsyncFileLoader` class. If this
            dictionary contains the ``pw_launch_kwargs`` key, it's value
            (assumes to be another dictionary) will be used to
            initialize the playwright instances used for the crawl.
            By default, ``None``.
        already_visited : set, optional
            A set of URLs (either strings or ``Link`` objects)
            that have already been visited. This is used to avoid
            re-visiting links that have already been checked.
            By default, ``None``.
        num_link_scores_to_check_per_page : int, default=3
            Number of top unique-scoring links per page to use for
            recursive crawling. This helps the crawl stay focused on the
            most likely links to contain documents of interest.
        max_pages : int, default=100
            Maximum number of pages to crawl before terminating,
            regardless of whether the document was found or not.
            By default, ``100``.
        browser_semaphore : :class:`asyncio.Semaphore`, optional
            Semaphore instance that can be used to limit the number of
            playwright browsers open concurrently. If ``None``, no
            limits are applied. By default, ``None``.
        """
        self.validator = validator
        self.url_scorer = url_scorer
        self.num_scores_to_check_per_page = num_link_scores_to_check_per_page
        self.checked_previously = already_visited or set()
        self.max_pages = max_pages

        file_loader_kwargs = file_loader_kwargs or {}
        flk = {"verify_ssl": False}
        flk.update(file_loader_kwargs or {})
        self.afl = AsyncFileLoader(**flk)
        self.pw_launch_kwargs = (
            file_loader_kwargs.get("pw_launch_kwargs") or {}
        )
        self.browser_semaphore = (
            AsyncExitStack()
            if browser_semaphore is None
            else browser_semaphore
        )

        self._out_docs = []
        self._already_visited = {}
        self._should_stop = None

    async def run(
        self, base_url, termination_callback=None, on_new_page_visit_hook=None
    ):
        """Run the COMPASS website crawler

        Parameters
        ----------
        base_url : str
            URL of the website to start crawling from.
        termination_callback : callable, optional
            An async callable that takes a list of documents and returns
            a boolean indicating whether to stop crawling. If ``None``,
            the crawl will simply terminates when :obj:`DOC_THRESHOLD`
            number of documents have been found. By default, ``None``.
        on_new_page_visit_hook : callable, optional
            An async callable that is called every time a new page is
            found during the crawl. The callable should accept a single
            argument, which is the page ``Link`` instance.
            If ``None``, no additional processing is done on new pages.
            By default, ``None``.

        Returns
        -------
        list
            List of document instances that passed the validation
            check. Each document contains the text from a PDF and has an
            attribute `source` that contains the URL of the document.
            This list may be empty if no documents are found.
        """
        self._should_stop = termination_callback or _default_found_enough_docs
        await self._run(
            base_url, on_new_page_visit_hook=on_new_page_visit_hook
        )
        self._should_stop = None

        self._log_crawl_stats()

        if self._out_docs:
            self._out_docs.sort(key=lambda x: -1 * x.attrs[_SCORE_KEY])

        return self._out_docs

    async def _run(
        self,
        base_url,
        link=None,
        depth=0,
        score=0,
        on_new_page_visit_hook=None,
    ):
        """Recursive web crawl function"""
        if link is None:
            base_url, link = self._reset_crawl(base_url)

        if link in self._already_visited:
            return

        if on_new_page_visit_hook:
            await on_new_page_visit_hook(link)

        self._already_visited[link] = (depth, score)
        logger.trace("self._already_visited=%r", self._already_visited)

        if await self._website_link_is_doc(link, depth, score):
            return

        num_urls_checked_on_this_page = 0
        curr_url_score = None
        for next_link in await self._get_links_from_page(link, base_url):
            prev_len = len(self._out_docs)
            await self._run(
                base_url,
                link=_Link(
                    title=next_link["title"],
                    href=next_link["href"],
                    base_domain=base_url,
                ),
                depth=depth + 1,
                score=next_link["score"],
                on_new_page_visit_hook=on_new_page_visit_hook,
            )

            doc_was_just_found = (  # fmt: off
                len(self._out_docs) == (prev_len + 1)
                and (
                    self._out_docs[-1].attrs.get(_DEPTH_KEY, -1) == (depth + 1)
                )
            )
            if doc_was_just_found:
                if await self.validator(self._out_docs[-1]):
                    logger.debug("    - Document passed validation check!")
                else:
                    self._out_docs = self._out_docs[:-1]
            elif (
                not link.resembles_pdf and curr_url_score != next_link["score"]
            ):
                logger.trace(
                    "Finished checking score %d at depth %d. Next score: %d",
                    curr_url_score or -1,
                    depth,
                    next_link["score"],
                )
                num_urls_checked_on_this_page += 1
                curr_url_score = next_link["score"]

            if await self._should_terminate_crawl(
                num_urls_checked_on_this_page, link
            ):
                break

        return

    def _reset_crawl(self, base_url):
        """Reset crawl state and initialize crawling link"""
        self._out_docs = []
        self._already_visited = {}
        base_url = _sanitize_url(base_url)
        return base_url, _Link(
            title="Landing Page",
            href=_sanitize_url(urljoin(base_url, base_url.split(" ")[0])),
            base_domain=base_url,
        )

    async def _website_link_is_doc(self, link, depth, score):
        """Check if website link contains doc"""
        if link in self.checked_previously and link.consistent_domain:
            # Don't re-check pages on main website
            return False

        if await self._website_link_is_pdf(link, depth, score):
            return True

        # at this point the page is NOT a PDF. However, it could still
        # just be a normal webpage on the main domain that we haven't
        # visited before. In that case, just return False
        if not link.consistent_domain:
            return False

        # now we are on an external page that we either have not visited
        # before or that we have seen but determined is NOT a PDF file
        return await self._website_link_as_html_doc(link, depth, score)

    async def _website_link_is_pdf(self, link, depth, score):
        """Fetch page content; keep only PDFs"""
        logger.debug("Loading Link: %s", link)

        try:
            doc = await self.afl.fetch(link.href)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            msg = (
                "Encountered error of type %r while trying to fetch "
                "content from %s"
            )
            err_type = type(e)
            logger.exception(msg, err_type, link)
            return False

        if isinstance(doc, PDFDocument):
            logger.debug("    - Found PDF!")
            doc.attrs[_DEPTH_KEY] = depth
            doc.attrs[_SCORE_KEY] = score
            self._out_docs.append(doc)
            return True

        return False

    async def _website_link_as_html_doc(self, link, depth, score):
        """Fetch page content as HTML doc"""
        logger.debug("Loading Link as HTML: %s", link)
        html_text = await self._get_text_no_err(link.href)

        attrs = {_DEPTH_KEY: depth, _SCORE_KEY: score}
        doc = HTMLDocument([html_text], attrs=attrs)
        self._out_docs.append(doc)
        return True

    async def _get_links_from_page(self, link, base_url):
        """Get all links from a page sorted by relevance score"""
        if not link.consistent_domain:
            logger.debug("Detected new domain, stopping link discovery")
            return []

        html_text = await self._get_text_no_err(link.href)
        page_links = []
        if html_text:
            page_links = _extract_links_from_html(html_text, base_url=base_url)
            page_links = await self.url_scorer(
                [dict(link) for link in page_links]
            )
            page_links = sorted(
                page_links, key=operator.itemgetter("score"), reverse=True
            )
        _debug_info_on_links(page_links)
        return page_links

    async def _get_text_no_err(self, url):
        """Get all text from a page; return empty string if pw error"""
        try:
            text = await self._get_text(url)
        except (PlaywrightError, RBPlaywrightError):
            text = ""

        return text

    async def _get_text(self, url):
        """Get all html text from a page"""
        all_text = []
        pw_page_kwargs = {
            "intercept_routes": True,
            "ignore_https_errors": True,
            "timeout": 60_0000,
        }
        async with async_playwright() as p, self.browser_semaphore:
            browser = await p.chromium.launch(**self.pw_launch_kwargs)
            async with pw_page(browser, **pw_page_kwargs) as page:
                await page.goto(url)
                await page.wait_for_load_state("networkidle", timeout=60_000)

                all_text.append(await page.content())
                all_text += await _get_text_from_all_locators(page)

        return "\n".join(all_text)

    async def _should_terminate_crawl(
        self, num_urls_checked_on_this_page, link
    ):
        """Check if crawl should terminate"""
        if num_urls_checked_on_this_page >= self.num_scores_to_check_per_page:
            logger.debug(
                "Already checked %d unique link scores from %s",
                self.num_scores_to_check_per_page,
                link.href,
            )
            return True

        if await self._should_stop(self._out_docs):
            logger.debug("Exiting crawl early due to user condition")
            return True

        if len(self._already_visited) >= self.max_pages:
            logger.debug("    - Too many links visited, stopping recursion")
            return True

        logger.trace(
            "Only checked %d pages, continuing crawl...",
            len(self._already_visited),
        )
        return False

    def _log_crawl_stats(self):
        """Log statistics about crawled pages and depths"""
        logger.info("Crawled %d pages", len(self._already_visited))
        logger.info("Found %d potential documents", len(self._out_docs))
        logger.debug("Average score: %.2f", self._compute_avg_link_score())

        logger.debug("Pages crawled by depth:")
        for depth, count in sorted(self._crawl_depth_counts().items()):
            logger.debug("  Depth %d: %d pages", depth, count)

    def _compute_avg_link_score(self):
        """Compute the average score of the crawled results"""
        return sum(
            score for __, score in self._already_visited.values()
        ) / len(self._already_visited)

    def _crawl_depth_counts(self):
        """Compute number of pages per depth"""
        depth_counts = Counter()
        depth_counts.update([d for d, __ in self._already_visited.values()])
        return depth_counts


async def _default_found_enough_docs(out_docs):  # noqa: RUF029
    """Check if a predetermined # of documents has been found

    The number to check is set by the module-level constant
    :obj:`DOC_THRESHOLD`.
    """
    return len(out_docs) >= DOC_THRESHOLD


def _debug_info_on_links(links):
    """Send debug info on links to logger"""
    num_links = len(links)
    if num_links <= 0:
        logger.debug("Found no links!")
        return

    logger.debug("Found %d links:", len(links))

    for link in links[:3]:
        logger.debug(
            "    - %d: %s (%s)", link["score"], link["title"], link["href"]
        )
    if num_links > 3:  # noqa: PLR2004
        logger.debug("    ...")


def _sanitize_url(url):
    """Fix common URL issues

    - Encode spaces and unsafe characters in the path
    - Encode query parameters safely
    - Leave existing percent-encoding intact
    """
    parsed = urlparse(url)

    safe_path = quote(unquote(parsed.path), safe="/")

    query_params = parse_qsl(parsed.query, keep_blank_values=True)
    safe_query = urlencode(query_params, doseq=True)  # cspell: disable-line

    return urlunparse(
        (
            parsed.scheme,
            parsed.netloc,
            safe_path,
            parsed.params,
            safe_query,
            parsed.fragment,
        )
    )


def _extract_links_from_html(text, base_url):
    """Parse HTML and extract all links"""
    soup = BeautifulSoup(text, "html.parser")
    links = [
        (a.get_text().strip(), a["href"])
        for a in soup.find_all("a", href=True)
    ]

    return {
        _Link(
            title=title,
            href=_sanitize_url(urljoin(base_url, path)),
            base_domain=base_url,
        )
        for (title, path) in links
        if title
        and path
        and not any(
            substr in title.lower() for substr in _BLACKLIST_SUBSTRINGS
        )
        and not any(substr in path.lower() for substr in _BLACKLIST_SUBSTRINGS)
    }


async def _get_text_from_all_locators(page):
    """Go through locators on page and get text behind them"""
    all_text = []
    for selector in _CLICKABLE_SELECTORS:
        logger.trace("Checking selector %r", selector)
        locators = page.locator(selector)
        locator_count = await locators.count()
        logger.trace("    - Found %d instances", locator_count)
        for index in range(locator_count):
            try:
                text = await _get_locator_text(locators, index, page)
            except (PlaywrightError, RBPlaywrightError):
                continue
            if text:
                all_text.append(text)
    return all_text


async def _get_locator_text(locators, index, page):
    """Get text after clicking on one of the page locators"""
    locator = locators.nth(index)

    if not await locator.is_visible():
        return None

    if not await locator.is_enabled():
        return None

    await locator.click(timeout=10_000)
    return await page.content()
